/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include "cidev.h"
#include "gfsconf.h"
#include "gfsops.h"
#include "gfs_release.cf"
#include "copyright.cf"
#include "libparser.h"
#include "upgrade.h"

char *prog_name;


int stat_config_file(char *filename)
{
    struct stat devstat;

    memset(&devstat, 0, sizeof(struct stat));

    if (stat(filename, &devstat) < 0)
        die("Error - cannot stat \"%s\"\n", filename);

    if (!S_ISREG(devstat.st_mode))
        die("Error - \"%s\" is not a regular file\n", filename);

    return 0;
}


int open_cidev(char *name)
{
    struct stat devstat;
    int cfd;

    memset(&devstat, 0, sizeof(struct stat));

    if (stat(name, &devstat) < 0)
        die("Error - cannot stat \"%s\"\n", name);

    if (!S_ISBLK(devstat.st_mode) && !S_ISCHR(devstat.st_mode))
        die("Error - \"%s\" is not a block device\n", name);

    cfd = open(name, O_RDWR);
    if (cfd < 0)
        die("Error - can't open \"%s\"\n", name);

    return cfd;
}


int do_cidev_read(int cfd, cidev_entry_t **gen_config)
{
    int count = 0, i, error, ret_count;
    cidev_entry_t *ent, *prev = NULL;

    if (cidev_entry_count(cfd, &count))
        return -1;

    ent = (cidev_entry_t *)malloc(count * sizeof(cidev_entry_t));
    if (!ent)
        return -ENOMEM;
    memset(ent, 0, count * sizeof(cidev_entry_t));
    *gen_config = ent;

    for (i = 0; i < count; i++)
    {
        if (prev)
        {
            ent++;
            prev->next = ent;
        }
        ent->next = NULL;
        prev = ent;
    }

    error = cidev_read_entries(cfd, *gen_config, &ret_count);
    if (error)
        return error;

    /* do something if ret_count != count ? */

    return 0;
}


static void usage(char *prog, FILE *file)
{
    fprintf(file, "Usage:\n\n");

    fprintf(file, "%s [options]\n\n", prog);

    fprintf(file, "Options:\n");

    fprintf(file, "  -d <cidev>         Specify the cidev to operate on\n");
    fprintf(file, "  -e                 Erase configuration on specified cidev\n"
);
    fprintf(file, "  -f <file>          File to use for test (-t)\n");
    fprintf(file, "  -h                 Help with usage\n");
    fprintf(file, "  -O                 Force\n");
    fprintf(file, "  -p                 Print configuration on specified cidev\n");
    fprintf(file, "  -q                 Quiet\n");
    fprintf(file, "  -t                 Parse configuration file and exit\n");
    fprintf(file, "  -u <cluster name>  Upgrade cidev to CCS files\n");
    fprintf(file, "  -V                 Version of program\n");
}

inline void mutually_exclusive(char *progname, char flag1, char flag2)
{
    fprintf(stderr, "%s: flag '%c' conflicts with '%c'.\n", progname, flag2,
            flag1);
    fprintf(stderr, "See man page for details.\n");
    exit(1);
}

int main(int argc, char **argv)
{
    cidev_entry_t *gen_config;
    cidev_header_t head;
    int opt, create = 0, cfd, force = 0, quiet = 0, erase = 0, print = 0, test =0, upgrade = 0; 
    int add = 0, remove = 0;
    char *conffile = NULL, *cidevfile = NULL;
    char *cluster_name = NULL;

    prog_name = argv[0];

    while ((opt = getopt(argc, argv, "thVcd:aepr:f:Oqu:")) != EOF)
    {
        switch (opt)
        {
        case 'h':
            usage(argv[0], stdout);
            exit(0);

        case 'V':
            printf("%s %s (built %s %s)\n", argv[0], GFS_RELEASE_NAME,
                   __DATE__, __TIME__);
            printf("%s\n", GFS_COPYRIGHT);
            exit(0);

        case 'd':
            cidevfile = optarg;
            break;

        case 'c':
            printf("Writing a CIDEV is no longer supported.\n");
            printf("See CCS(8) for information on configuring GFS.\n");
            exit(1);
            break;

        case 'p':
            if(add)
                mutually_exclusive(argv[0], opt, 'a');
            if(erase)
                mutually_exclusive(argv[0], opt, 'e');
            if(create)
                mutually_exclusive(argv[0], opt, 'c');
            if(conffile)
                mutually_exclusive(argv[0], opt, 'f');
            if(remove)
                mutually_exclusive(argv[0], opt, 'r');

            print++;
            break;

        case 'e':
            if(add)
                mutually_exclusive(argv[0], opt, 'a');
            if(create)
                mutually_exclusive(argv[0], opt, 'c');
            if(print)
                mutually_exclusive(argv[0], opt, 'p');
            if(conffile)
                mutually_exclusive(argv[0], opt, 'f');
            if(remove)
                mutually_exclusive(argv[0], opt, 'r');

            erase++;
            break;

        case 'r':
            printf("Writing a CIDEV is no longer supported.\n");
            printf("See CCS(8) for information on configuring GFS.\n");
            exit(1);
            break;

        case 'a':
            printf("Writing a CIDEV is no longer supported.\n");
            printf("See CCS(8) for information on configuring GFS.\n");
            exit(1);
            break;

        case 'f':
            if(erase)
                mutually_exclusive(argv[0], opt, 'e');
            if(print)
                mutually_exclusive(argv[0], opt, 'p');
            if(remove)
                mutually_exclusive(argv[0], opt, 'f');

            if (strcmp(optarg,"-"))
                conffile = optarg;

           break;

        case 'q':
            quiet++;
            break;

        case 't':
            test++;
            break;

	case 'u':
            if(add)
                mutually_exclusive(argv[0], opt, 'a');
            if(create)
                mutually_exclusive(argv[0], opt, 'c');
            if(print)
                mutually_exclusive(argv[0], opt, 'p');
            if(test)
                mutually_exclusive(argv[0], opt, 't');
            if(remove)
                mutually_exclusive(argv[0], opt, 'r');
            if(erase)
                mutually_exclusive(argv[0], opt, 'e');
	    upgrade++;
	    cluster_name = optarg;
	    if(cluster_name[0] == '-'){
	      fprintf(stderr, "%s: option requires an argument -- u\n",
		      argv[0]);
	      fprintf(stderr, "Please use '-h' for usage.\n");
	      exit(1);
	    }
	    break;
        case 'O':
            force++;
            break;

        case ':':
        case '?':
            fprintf(stderr, "Please use '-h' for usage.\n");
            exit(1);
            break;

        default:
          fprintf(stderr, "Bad programmer! You forgot to catch the %c flag\n",
                  opt);
          exit(1);
          break;

        }
    }

    if (test)
    {
        if (conffile)
            stat_config_file(conffile);

        if (read_config_file(conffile, &gen_config))
            die("syntax error\n");

	return 0;
    }

    if(upgrade){
      /* specify one and only one */
      /* eventually also take cidev... 
      if((!cidevfile && !conffile) || (cidevfile && conffile)){
	fprintf(stderr, "'-u' requires one and only one '-f' or '-d' flag.\n");
	exit(1);
      }
      */
      if(!conffile){
	fprintf(stderr, "'-u' requires the use of the '-f' option.\n");
	exit(1);
      }
      if(cidev_to_ccs(conffile, cluster_name)){
	fprintf(stderr, "Upgrade failed.\n");
	exit(1);
      }
      printf("New Cluster Configuration System (CCS) files "
	     "are located in \"%s\"\n", cluster_name);
      printf("\nYou will need to do the following to finish your upgrade:\n"
	     "1) Look over the generated CCS files for completeness.\n"
	     "2) Add your license file (license.ccs) to the directory \"%s\"\n"
	     , cluster_name);

      exit(0);
    }
    if(!cidevfile ) {
        fprintf(stderr, "No '-d' flag specified.\n");
        fprintf(stderr, "Please use '-h' for usage.\n");
        exit(1);
    }

    if (erase)
    {
        cfd = open_cidev(cidevfile);

        if (cidev_read_header(cfd, &head))
            die("Bad cidev header\n");

        if (head.magic != CIDEV_MAGIC)
            die("no formatted cidev to erase\n");

        if (!force && ((head.version & 0xFFFFFF00) != (CIDEV_VERSION & 0xFFFFFF00)))
            die("cidev version mismatch -  see force option\n");

        if (cidev_erase(cfd))
            return -1;

        if (!quiet)
            printf("%s erased.\n", cidevfile);
        return 0;
    }

    if (print)
    {
        cfd = open_cidev(cidevfile);

        if (cidev_read_header(cfd, &head))
            die("Bad cidev header\n");

        if (!force && (head.magic != CIDEV_MAGIC))
            die("not a valid cidev\n");

        if (!force && ((head.version & 0xFFFFFF00) != (CIDEV_VERSION & 0xFFFFFF00)))
            die("cidev version mismatch - see force option\n");

        if (do_cidev_read(cfd, &gen_config))
            return -1;

        print_config_file(gen_config);
        return 0;
    }

    fprintf(stderr, "You must specify one of the `-acepr' flags.\n");
    fprintf(stderr, "Use the '-h' flag for usage.\n");
    return 1;
}

