/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include <linux/posix_acl.h>

#include "gfs.h"
#include "acl.h"
#include "bmap.h"
#include "dio.h"
#include "dir.h"
#include "eattr.h"
#include "glock.h"
#include "glops.h"
#include "inode.h"
#include "log.h"
#include "quota.h"
#include "rgrp.h"
#include "super.h"
#include "trans.h"
#include "unlinked.h"

/**
 * make_dinode - Fill in a new dinode structure
 * @dip: the directory this inode is being created in
 * @gl: The glock covering the new inode
 * @inum: the inode number
 * @type: the file type
 * @mode: the file permissions
 * @cred: a credentials structure
 *
 */

static int
make_dinode(struct gfs_inode *dip, struct gfs_glock *gl, struct gfs_inum *inum,
	    unsigned int type, unsigned int mode, int *cred)
{
	struct gfs_sbd *sdp = dip->i_sbd;
	struct gfs_dinode di;
	struct buffer_head *dibh;
	struct gfs_rgrpd *rgd;
	int error;

	error =
	    gfs_dread(sdp, inum->no_addr, gl, DIO_NEW | DIO_START | DIO_WAIT,
		      &dibh);
	if (error)
		goto out;

	gfs_trans_add_bh(sdp, gl, dibh);
	gfs_metatype_set(sdp, dibh, GFS_METATYPE_DI, GFS_FORMAT_DI);
	gfs_buffer_clear_tail(dibh, sizeof (struct gfs_dinode));

	memset(&di, 0, sizeof (struct gfs_dinode));

	gfs_meta_header_in(&di.di_header, (dibh)->b_data);

	di.di_num = *inum;

	if (dip->i_di.di_mode & 02000) {
		di.di_mode = mode | ((type == GFS_FILE_DIR) ? 02000 : 0);
		di.di_gid = dip->i_di.di_gid;
	} else {
		di.di_mode = mode;
		di.di_gid = ((cred) ? current->fsgid : 0);
	}

	di.di_uid = ((cred) ? current->fsuid : 0);
	di.di_nlink = 1;
	di.di_blocks = 1;
	di.di_atime = di.di_mtime = di.di_ctime = CURRENT_TIME;

	rgd = gfs_blk2rgrpd(sdp, inum->no_addr);
	GFS_ASSERT_SBD(rgd, sdp, printk("block = %" PRIu64 "\n", inum->no_addr);
	    );

	di.di_rgrp = rgd->rd_ri.ri_addr;
	di.di_goal_rgrp = di.di_rgrp;
	di.di_goal_dblk = di.di_goal_mblk = inum->no_addr - rgd->rd_ri.ri_data1;

	if (type == GFS_FILE_REG) {
		if ((dip->i_di.di_flags & GFS_DIF_INHERIT_JDATA)
		    || sdp->sd_tune.gt_new_files_jdata)
			di.di_flags |= GFS_DIF_JDATA;
		if ((dip->i_di.di_flags & GFS_DIF_INHERIT_DIRECTIO)
		    || sdp->sd_tune.gt_new_files_directio)
			di.di_flags |= GFS_DIF_DIRECTIO;
	} else if (type == GFS_FILE_DIR) {
		di.di_flags |= (dip->i_di.di_flags & GFS_DIF_INHERIT_DIRECTIO);
		di.di_flags |= (dip->i_di.di_flags & GFS_DIF_INHERIT_JDATA);
	}

	di.di_type = type;

	gfs_dinode_out(&di, (dibh)->b_data);
	gfs_drelse(dibh);

	gfs_trans_add_quota(sdp, +1, di.di_uid, di.di_gid);

      out:
	return error;
}

/**
 * gfs_copyin_dinode - Refresh the incore copy of the dinode
 * @ip: The GFS inode
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
gfs_copyin_dinode(struct gfs_inode *ip)
{
	struct buffer_head *dibh;
	int error;

	error = gfs_get_inode_buffer(ip, &dibh);
	if (error)
		return error;

	gfs_metatype_check(ip->i_sbd, dibh, GFS_METATYPE_DI);
	gfs_dinode_in(&ip->i_di, (dibh)->b_data);

	gfs_drelse(dibh);

	GFS_ASSERT_INODE(ip->i_num.no_formal_ino ==
			 ip->i_di.di_num.no_formal_ino, ip,
			 gfs_dinode_print(&ip->i_di, TRUE);
	    );

	/*  Handle a moved inode  */

	if (ip->i_num.no_addr != ip->i_di.di_num.no_addr) {
		GFS_ASSERT_INODE(FALSE, ip,);	/*  Not implemented yet  */
	}

	ip->i_vn = ip->i_gl->gl_vn;

	return 0;
}

/**
 * create_istruct - create a struct gfs_inode
 * @i_gl: The glock covering the inode
 * @inum: The inode number
 * @io_gl: the iopen glock, or NULL
 * @io_state: the state the iopen glock should be acquire in
 * @ipp: pointer to put the returned inode in
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
create_istruct(struct gfs_glock *i_gl, struct gfs_inum *inum,
	       struct gfs_glock *io_gl, unsigned int io_state,
	       struct gfs_inode **ipp)
{
	struct gfs_sbd *sdp = i_gl->gl_sbd;
	struct gfs_inode *ip;
	int error = 0;

	ip = gmalloc_inode();

	ip->i_num = *inum;

	atomic_set(&ip->i_count, 1);

	ip->i_gl = i_gl;
	ip->i_sbd = sdp;

	error = gfs_glock_nq_init(io_gl,
				  io_state, GL_LOCAL_EXCL | GL_EXACT,
				  &ip->i_iopen_gh);
	if (error)
		goto fail;

	ip->i_iopen_gh.gh_owner = NULL;	/*  Prevent recursion later  */

	spin_lock(&io_gl->gl_spin);
	gfs_hold_glstruct(i_gl);
	io_gl->gl_object = i_gl;
	spin_unlock(&io_gl->gl_spin);

	error = gfs_copyin_dinode(ip);
	if (error)
		goto fail_iopen;

	gfs_hold_glstruct(i_gl);
	i_gl->gl_object = ip;

	atomic_inc(&sdp->sd_inode_count);

	*ipp = ip;

	return 0;

      fail_iopen:
	spin_lock(&io_gl->gl_spin);
	io_gl->gl_object = NULL;
	gfs_put_glstruct(i_gl);
	spin_unlock(&io_gl->gl_spin);

	ip->i_iopen_gh.gh_flags |= GL_NOCACHE;
	gfs_glock_dq_uninit(&ip->i_iopen_gh);

      fail:
	gfree_inode(ip);
	*ipp = NULL;

	return error;
}

/**
 * gfs_get_istruct - Get an inode given its number
 * @i_gl: The glock covering the inode
 * @inum: The inode number
 * @create: Flag to say if we are allowed to create a new struct gfs_inode
 * @ipp: pointer to put the returned inode in
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
gfs_get_istruct(struct gfs_glock *i_gl, struct gfs_inum *inum, int create,
		struct gfs_inode **ipp)
{
	struct gfs_glock *io_gl;
	int error = 0;

	*ipp = (struct gfs_inode *) i_gl->gl_object;
	if (*ipp) {
		atomic_inc(&(*ipp)->i_count);
		GFS_ASSERT_INODE((*ipp)->i_num.no_formal_ino ==
				 inum->no_formal_ino, (*ipp),);
	} else if (create) {
		error = gfs_get_glstruct(i_gl->gl_sbd,
					 inum->no_addr, &gfs_iopen_glops,
					 ST_CREATE, &io_gl);
		if (!error) {
			error =
			    create_istruct(i_gl, inum, io_gl, LM_ST_SHARED,
					   ipp);
			gfs_put_glstruct(io_gl);
		}
	}

	return error;
}

/**
 * gfs_hold_istruct - hold a struct gfs_inode structure
 * @ip: The GFS inode
 *
 */

void
gfs_hold_istruct(struct gfs_inode *ip)
{
	GFS_ASSERT_INODE(atomic_read(&ip->i_count), ip,);
	atomic_inc(&ip->i_count);
	return;
}

/**
 * gfs_put_istruct - put a struct gfs_inode structure
 * @ip: The GFS inode
 *
 */

void
gfs_put_istruct(struct gfs_inode *ip)
{
	GFS_ASSERT_INODE(atomic_read(&ip->i_count) > 0, ip,);
	atomic_dec(&ip->i_count);
	return;
}

/**
 * gfs_destroy_istruct - Destroy an inode structure with no references on it
 * @ip: The GFS inode
 *
 * This function must be called with a glock held on the inode.
 *
 */

void
gfs_destroy_istruct(struct gfs_inode *ip)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_glock *io_gl = ip->i_iopen_gh.gh_gl;
	struct gfs_glock *i_gl = ip->i_gl;

	GFS_ASSERT_INODE(!atomic_read(&ip->i_count), ip,);
	GFS_ASSERT_INODE((struct gfs_glock *) io_gl->gl_object == i_gl, ip,);

	spin_lock(&io_gl->gl_spin);
	io_gl->gl_object = NULL;
	gfs_put_glstruct(i_gl);
	spin_unlock(&io_gl->gl_spin);

	ip->i_iopen_gh.gh_flags |= GL_NOCACHE;
	gfs_glock_dq_uninit(&ip->i_iopen_gh);

	gfree_inode(ip);

	i_gl->gl_object = NULL;
	gfs_put_glstruct(i_gl);

	atomic_dec(&sdp->sd_inode_count);

	return;
}

/**
 * dinode_dealloc - Put deallocate a dinode
 * @ip: The GFS inode
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
dinode_dealloc(struct gfs_inode *ip)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_rgrpd *rgd;
	struct gfs_holder ri_gh, rgd_gh;
	int error;

	gfs_get_alloc(ip);

	error = gfs_hold_quota_structs(ip, NO_QUOTA_CHANGE, NO_QUOTA_CHANGE);
	if (error)
		goto fail;

	error = gfs_rindex_hold(sdp, &ri_gh);
	if (error)
		goto fail_qs;

	rgd = gfs_blk2rgrpd(sdp, ip->i_num.no_addr);
	GFS_ASSERT_INODE(rgd, ip,);

	error = gfs_glock_nq_init(rgd->rd_gl, LM_ST_EXCLUSIVE, 0, &rgd_gh);
	if (error)
		goto fail_rindex_relse;

	GFS_ASSERT_INODE(ip->i_di.di_blocks == 1, ip,
			 gfs_dinode_print(&ip->i_di, TRUE);
	    );

	/*  Trans may require:
	   One block for the RG header.
	   One block for the dinode bit.
	   One block for the dinode.
	   We also need a block for the unlinked change.
	   One block for the quota change.  */

	error = gfs_trans_begin(sdp, TRANS_I_DEALLOC, 3, 2);
	if (error)
		goto fail_rg_gunlock;

	error = gfs_difree(ip, rgd);
	if (error)
		goto fail_end_trans;

	gfs_trans_add_unlinked(sdp, GFS_LOG_DESC_IDA, &ip->i_num);
	clear_bit(GLF_STICKY, &ip->i_gl->gl_flags);

	gfs_trans_end(sdp);

	gfs_glock_dq_uninit(&rgd_gh);

	gfs_glock_dq_uninit(&ri_gh);

	gfs_unhold_quota_structs(ip);
	gfs_put_alloc(ip);

	return 0;

      fail_end_trans:
	gfs_trans_end(sdp);

      fail_rg_gunlock:
	gfs_glock_dq_uninit(&rgd_gh);

      fail_rindex_relse:
	gfs_glock_dq_uninit(&ri_gh);

      fail_qs:
	gfs_unhold_quota_structs(ip);

      fail:
	gfs_put_alloc(ip);

	return error;
}

/**
 * inode_dealloc - Deallocate an inode
 * @sdp: the filesystem
 * @inum: the inode number to deallocate
 * @io_gh: a holder for the iopen glock for this inode
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
inode_dealloc(struct gfs_sbd *sdp, struct gfs_inum *inum,
	      struct gfs_holder *io_gh)
{
	struct gfs_inode *ip;
	struct gfs_holder i_gh;
	int error;

	error = gfs_glock_nq_num(sdp,
				 inum->no_formal_ino, &gfs_inode_glops,
				 LM_ST_EXCLUSIVE, 0, &i_gh);
	if (error)
		return error;

	/*  We reacquire the iopen lock here to avoid a race with the NFS server
	   calling gfs_read_inode() with the inode number of a inode we're in the
	   process of deallocating.  And we can't keep our hold on the lock
	   from try_dealloc_inode() for deadlock reasons.  */

	gfs_reinit_holder(LM_ST_EXCLUSIVE, LM_FLAG_TRY, io_gh);
	error = gfs_glock_nq(io_gh);
	switch (error) {
	case 0:
		break;
	case GLR_TRYFAILED:
		error = 0;
		goto fail;
	default:
		GFS_ASSERT_SBD(error < 0, sdp,);
		goto fail;
	}

	GFS_ASSERT_GLOCK(!i_gh.gh_gl->gl_object, i_gh.gh_gl,);
	error =
	    create_istruct(i_gh.gh_gl, inum, io_gh->gh_gl, LM_ST_EXCLUSIVE,
			   &ip);

	gfs_glock_dq(io_gh);

	if (error)
		goto fail;

	GFS_ASSERT_INODE(!ip->i_di.di_nlink, ip,
			 gfs_dinode_print(&ip->i_di, TRUE);
	    );
	GFS_ASSERT_INODE(atomic_read(&ip->i_count) == 1, ip,);
	GFS_ASSERT_INODE(!ip->i_vnode, ip,);

	if (ip->i_di.di_type == GFS_FILE_DIR
	    && (ip->i_di.di_flags & GFS_DIF_EXHASH)) {
		error = gfs_dir_exhash_free(ip);
		if (error)
			goto fail_iput;
	}

	if (ip->i_di.di_eattr) {
		error = gfs_ea_dealloc(ip);
		if (error)
			goto fail_iput;
	}

	error = gfs_shrink(ip, 0, NULL);
	if (error)
		goto fail_iput;

	error = dinode_dealloc(ip);
	if (error)
		goto fail_iput;

	gfs_put_istruct(ip);
	gfs_destroy_istruct(ip);

	gfs_glock_dq_uninit(&i_gh);

	return 0;

      fail_iput:
	gfs_put_istruct(ip);
	gfs_destroy_istruct(ip);

      fail:
	gfs_glock_dq_uninit(&i_gh);

	return error;
}

/**
 * gfs_try_dealloc_inode - Grab an unlinked inode off the list and try to free it.
 * @sdp: the filesystem
 *
 * Returns: 0 on successful deallocation, 1 on stuck inode,
 *          2 on no inodes, -EXXX on failure
 */

int
gfs_try_dealloc_inode(struct gfs_sbd *sdp)
{
	struct gfs_unlinked *ul;
	struct gfs_holder io_gh;
	int error = 0;

	ul = gfs_find_ulstruct(sdp);
	if (!ul)
		return 2;

	gfs_try_toss_inode(sdp, &ul->ul_inum);

	error = gfs_glock_nq_num(sdp,
				 ul->ul_inum.no_addr, &gfs_iopen_glops,
				 LM_ST_EXCLUSIVE, LM_FLAG_TRY_1CB, &io_gh);
	switch (error) {
	case 0:
		break;
	case GLR_TRYFAILED:
		error = 1;
		goto out;
	default:
		GFS_ASSERT_SBD(error < 0, sdp,);
		goto out;
	}

	gfs_glock_dq(&io_gh);
	error = inode_dealloc(sdp, &ul->ul_inum, &io_gh);
	gfs_uninit_holder(&io_gh);

      out:
	gfs_unlock_ulstruct(sdp, ul);

	return error;
}

/**
 * gfs_limit_unlinked - limit the number of inodes waiting to be deallocated
 * @sdp: the filesystem
 *
 * Returns: 0 on success, -EXXX on failure;
 */

void
gfs_limit_unlinked(struct gfs_sbd *sdp)
{
	unsigned int tries = 0, min = 0;
	int error;

	if (atomic_read(&sdp->sd_unlinked_ic_count) >= sdp->sd_tune.gt_ilimit2) {
		tries = sdp->sd_tune.gt_ilimit2_tries;
		min = sdp->sd_tune.gt_ilimit2_min;
	} else if (atomic_read(&sdp->sd_unlinked_ic_count) >=
		   sdp->sd_tune.gt_ilimit1) {
		tries = sdp->sd_tune.gt_ilimit1_tries;
		min = sdp->sd_tune.gt_ilimit1_min;
	}

	while (tries--) {
		error = gfs_try_dealloc_inode(sdp);
		if (!error) {
			if (!--min)
				break;
		} else if (error != 1)
			break;
	}

}

/**
 * gfs_change_nlink - Change nlink count on inode
 * @ip: The GFS inode
 * @diff: The change in the nlink count required
 *
 * Returns: 0 on success, -EXXXX on failure.
 */

int
gfs_change_nlink(struct gfs_inode *ip, int diff)
{
	struct buffer_head *dibh;
	uint32_t nlink;
	int error;

	nlink = ip->i_di.di_nlink + diff;

	if (diff < 0)
		GFS_ASSERT_INODE(nlink < ip->i_di.di_nlink, ip,
				 gfs_dinode_print(&ip->i_di, TRUE);
	    );

	error = gfs_get_inode_buffer(ip, &dibh);
	if (error)
		goto out;

	ip->i_di.di_nlink = nlink;
	ip->i_di.di_ctime = CURRENT_TIME;

	gfs_trans_add_bh(ip->i_sbd, ip->i_gl, dibh);
	gfs_dinode_out(&ip->i_di, (dibh)->b_data);
	gfs_drelse(dibh);

      out:

	return error;
}

/**
 * gfs_lookupi - Look up a filename in a directory and return its inode
 * @d_gh: An initialized holder for the directory glock
 * @name: The name of the inode to look for
 * @cred: The caller's credentials
 * @i_gh: An uninitialized holder for the new inode glock
 *
 * Returns: 0 on success, -EXXXX on failure
 */

int
gfs_lookupi(struct gfs_holder *d_gh, struct qstr *name,
	    int *cred, struct gfs_holder *i_gh)
{
	struct gfs_inode *dip = (struct gfs_inode *) d_gh->gh_gl->gl_object;
	struct gfs_sbd *sdp = dip->i_sbd;
	struct gfs_glock *gl;
	struct gfs_inode *ip;
	struct gfs_inum inum, inum2;
	unsigned int type;
	struct buffer_head *dibh;
	struct gfs_dinode *di;
	int error;

	i_gh->gh_gl = NULL;

	if (!name->len || name->len > GFS_FNAMESIZE) {
		error = -ENAMETOOLONG;
		goto out;
	}

	if (gfs_filecmp(name, (char *) ".", 1)) {
		gfs_reinit_holder(LM_ST_SHARED, 0, d_gh);
		error = gfs_glock_nq(d_gh);
		if (!error) {
			error =
			    gfs_glock_nq_init(dip->i_gl, LM_ST_SHARED, 0, i_gh);
			GFS_ASSERT_INODE(!error, ip,);
			gfs_hold_istruct(dip);
		}

		goto out;
	}

	if (gfs_glock_is_locked_by_me(d_gh->gh_gl))
		bitch_about(sdp, &sdp->sd_last_readdirplus,
			    "readdirplus-type behavior");

 restart1:
	gfs_reinit_holder(LM_ST_SHARED, 0, d_gh);
	error = gfs_glock_nq(d_gh);
	if (error)
		goto out;

	error = gfs_permission(dip, GFS_MAY_EXEC, cred);
	if (error) {
		gfs_glock_dq(d_gh);
		goto out;
	}

	error = gfs_dir_search(dip, name, &inum, &type);
	if (error) {
		gfs_glock_dq(d_gh);
		if (error == -ENOENT)
			error = 0;
		goto out;
	}

 restart2:
	error = gfs_get_glstruct(sdp, inum.no_formal_ino, &gfs_inode_glops,
				 ST_CREATE, &gl);
	if (error) {
		gfs_glock_dq(d_gh);
		goto out;
	}

	/*  Acquire the second lock  */

	if (gl->gl_name.ln_number < dip->i_gl->gl_name.ln_number) {
		gfs_glock_dq(d_gh);

		error = gfs_glock_nq_init(gl, LM_ST_SHARED,
					  LM_FLAG_ANY | GL_LOCAL_EXCL, i_gh);
		if (error)
			goto out_glput;

		error = gfs_dread(sdp, inum.no_addr, gl, DIO_START | DIO_WAIT, &dibh);
		if (error)
		{
			gfs_glock_dq_uninit(i_gh);
			goto out_glput;
		}

		di = (struct gfs_dinode *)dibh->b_data;

		if (gfs32_to_cpu(di->di_header.mh_magic) != GFS_MAGIC ||
		    gfs32_to_cpu(di->di_header.mh_type) != GFS_METATYPE_DI ||
		    (gfs32_to_cpu(di->di_flags) & GFS_DIF_UNUSED) ||
		    !di->di_nlink)
		{
			gfs_drelse(dibh);
			gfs_glock_dq_uninit(i_gh);
			gfs_put_glstruct(gl);
			goto restart1;
		}

		gfs_drelse(dibh);

		gfs_reinit_holder(LM_ST_SHARED, 0, d_gh);
		error = gfs_glock_nq(d_gh);
		if (error) {
			gfs_glock_dq_uninit(i_gh);
			goto out_glput;
		}

		error = gfs_permission(dip, GFS_MAY_EXEC, cred);
		if (error) {
			gfs_glock_dq(d_gh);
			gfs_glock_dq_uninit(i_gh);
			goto out_glput;
		}

		error = gfs_dir_search(dip, name, &inum2, &type);
		if (error) {
			gfs_glock_dq(d_gh);
			gfs_glock_dq_uninit(i_gh);
			if (error == -ENOENT)
				error = 0;
			goto out_glput;
		}

		if (!GFS_INUM_EQUAL(&inum, &inum2)) {
			gfs_glock_dq_uninit(i_gh);
			gfs_put_glstruct(gl);
			inum = inum2;
			goto restart2;
		}
	} else {
		error =
		    gfs_glock_nq_init(gl, LM_ST_SHARED,
				      LM_FLAG_ANY | GL_LOCAL_EXCL, i_gh);
		if (error) {
			gfs_glock_dq(d_gh);
			goto out_glput;
		}
	}

	error = gfs_get_istruct(gl, &inum, ST_CREATE, &ip);
	if (error) {
		gfs_glock_dq(d_gh);
		gfs_glock_dq_uninit(i_gh);
	}
	GFS_ASSERT_INODE(ip == i_gh->gh_gl->gl_object, ip,);
	GFS_ASSERT_INODE(ip->i_di.di_type == type, ip,);

      out_glput:
	gfs_put_glstruct(gl);

      out:

	return error;
}

/**
 * pick_formal_ino - Pick a formal inode number for a given inode
 * @sdp: the filesystem
 * @inum: the inode number structure
 *
 */

static void
pick_formal_ino(struct gfs_sbd *sdp, struct gfs_inum *inum)
{

	/*  This won't always be true  */
	inum->no_formal_ino = inum->no_addr;

}

/**
 * gfs_createi - Create a new inode
 * @d_gh: An initialized holder for the directory glock
 * @name: The name of the new file
 * @type: The type of dinode (GFS_FILE_REG, GFS_FILE_DIR, GFS_FILE_LNK, ...)
 * @mode: the permissions on the new inode
 * @cred: the caller's credentials
 * @new: If NULL, create is exclusive, else set *new to TRUE if the inode is new
 * @i_gh: An uninitialized holder for the new inode glock
 *
 * If the return value is 0, the glocks on both the directory and the new
 * file are held.  A transaction has been started and an inplace reservation
 * is held, as well.
 *
 * Returns: 0 on success, -EXXXX on failure
 */

int
gfs_createi(struct gfs_holder *d_gh, struct qstr *name,
	    unsigned int type, unsigned int mode, int *cred,
	    int *new, struct gfs_holder *i_gh)
{
	struct gfs_inode *dip = (struct gfs_inode *) d_gh->gh_gl->gl_object;
	struct gfs_sbd *sdp = dip->i_sbd;
	struct gfs_alloc *al;
	struct gfs_holder io_gh;
	struct gfs_inode *ip;
	struct gfs_inum inum;
	int alloc_required;
	int error;
	struct posix_acl *acl = NULL;

	if (!name->len || name->len > GFS_FNAMESIZE) {
		error = -ENAMETOOLONG;
		goto fail;
	}

      restart:

	gfs_reinit_holder(LM_ST_EXCLUSIVE, 0, d_gh);
	error = gfs_glock_nq(d_gh);
	if (error)
		goto fail;

	error = gfs_permission(dip, GFS_MAY_WRITE | GFS_MAY_EXEC, cred);
	if (error)
		goto fail_gunlock;

	/*  Don't create entries in an unlinked directory  */

	if (!dip->i_di.di_nlink) {
		error = -EPERM;
		goto fail_gunlock;
	}

	error = gfs_dir_search(dip, name, NULL, NULL);
	switch (error) {
	case -ENOENT:
		break;

	case 0:
		if (!new) {
			error = -EEXIST;
			goto fail_gunlock;
		} else {
			gfs_glock_dq(d_gh);

			error = gfs_lookupi(d_gh, name, cred, i_gh);
			if (error)
				goto fail;

			if (i_gh->gh_gl) {
				*new = FALSE;
				return 0;
			} else
				goto restart;
		}

		break;

	default:
		goto fail_gunlock;
	}

	if (dip->i_di.di_entries == (uint32_t) - 1) {
		error = -EFBIG;
		goto fail_gunlock;
	}
	if (type == GFS_FILE_DIR && dip->i_di.di_nlink == (uint32_t) - 1) {
		error = -EMLINK;
		goto fail_gunlock;
	}

	error = gfs_setup_new_acl(dip, &mode, type, &acl);
	if (error)
		goto fail_gunlock;

	al = gfs_get_alloc(dip);

	error =
	    gfs_get_quota_locks(dip, ((cred) ? current->fsuid : 0),
				((cred) ? current->fsgid : 0), cred);
	if (error)
		goto fail_alloc;

	al->al_diblks = 1;
	al->al_metablks = sdp->sd_max_dirres + GFS_MAX_EA_ACL_BLKS;
	al->al_trans = TRANS_CREATE;

	error = gfs_inplace_reserv(dip);
	if (error)
		goto fail_gunlock_q;

	gfs_dipick(dip, &inum);

	if (!inum.no_formal_ino)
		pick_formal_ino(sdp, &inum);

	error = gfs_glock_nq_num(sdp,
				 inum.no_formal_ino, &gfs_inode_glops,
				 LM_ST_EXCLUSIVE, GL_SKIP, i_gh);
	if (error)
		goto fail_ipres;

	error = gfs_glock_nq_num(sdp,
				 inum.no_addr, &gfs_iopen_glops,
				 LM_ST_SHARED, GL_LOCAL_EXCL | GL_EXACT,
				 &io_gh);
	if (error)
		goto fail_gunlock2;

	error = gfs_diradd_alloc_required(dip, name, &alloc_required);
	if (error)
		goto fail_gunlock3;

	if (alloc_required) {
		error =
		    gfs_check_quota(dip, dip->i_di.di_uid, dip->i_di.di_gid);
		if (error)
			goto fail_gunlock3;
	}
	error =
	    gfs_check_quota(dip, ((cred) ? current->fsuid : 0),
			    ((cred)) ? current->fsgid : 0);
	if (error)
		goto fail_gunlock3;

	/*  Trans may require:
	   blocks for two dinodes, the directory blocks necessary for
	   a new entry, RG bitmap blocks for an allocation,
	   and one block for a quota change.  */

	error =
	    gfs_trans_begin(sdp, TRANS_CREATE,
			    2 + sdp->sd_max_dirres + GFS_MAX_EA_ACL_BLKS +
			    al->al_rgd->rd_ri.ri_length, 1);
	if (error)
		goto fail_gunlock3;

	error = gfs_dialloc(dip, &inum, i_gh->gh_gl);
	if (error)
		goto fail_end_trans;

	error = gfs_dir_add(dip, name, &inum, type);
	if (error)
		goto fail_end_trans;

	error = make_dinode(dip, i_gh->gh_gl, &inum, type, mode, cred);
	if (error)
		goto fail_end_trans;

	error = gfs_get_istruct(i_gh->gh_gl, &inum, ST_CREATE, &ip);
	if (error)
		goto fail_end_trans;
	GFS_ASSERT_INODE(ip == i_gh->gh_gl->gl_object, ip,);

	if (acl != NULL) {
		error = gfs_init_acl(dip, ip, type, acl);
		if (error)
			goto fail_end_trans;
	}

	gfs_glock_dq_uninit(&io_gh);

	if (new)
		*new = TRUE;

	return 0;

      fail_end_trans:
	gfs_trans_end(sdp);

      fail_gunlock3:
	gfs_glock_dq_uninit(&io_gh);

      fail_gunlock2:
	gfs_glock_dq_uninit(i_gh);

      fail_ipres:
	gfs_inplace_relse(dip);

      fail_gunlock_q:
	gfs_put_quota_locks(dip);

      fail_alloc:
	gfs_put_alloc(dip);

      fail_gunlock:
	gfs_glock_dq(d_gh);

      fail:
	return error;
}

/**
 * gfs_unlinki - Unlink a file
 * @dip: The inode of the directory
 * @name: The name of the file to be unlinked
 * @ip: The inode of the file to be removed
 *
 * Assumes Glocks on both dip and ip are held.
 *
 * Returns: 0 on success, -EXXXX on failure
 */

int
gfs_unlinki(struct gfs_inode *dip, struct qstr *name, struct gfs_inode *ip)
{
	struct gfs_sbd *sdp = dip->i_sbd;
	int error = 0;

	error = gfs_dir_del(dip, name);
	if (error)
		goto out;

	error = gfs_change_nlink(ip, -1);
	if (error)
		goto out;

	/*  If this inode is being unlinked from the directory structure,
	   we need to mark that in the log so that it isn't lost during
	   a crash.  */

	if (!ip->i_di.di_nlink) {
		gfs_trans_add_unlinked(sdp, GFS_LOG_DESC_IUL, &ip->i_num);
		set_bit(GLF_STICKY, &ip->i_gl->gl_flags);
	}

      out:

	return error;
}

/**
 * gfs_rmdiri - Remove a directory
 * @dip: The parent directory of the directory to be removed
 * @name: The name of the directory to be removed
 * @ip: The GFS inode of the directory to be removed
 *
 * Assumes Glocks on dip and ip are held
 *
 * Returns: 0 on success, -EXXXX on failure
 */

int
gfs_rmdiri(struct gfs_inode *dip, struct qstr *name, struct gfs_inode *ip)
{
	struct gfs_sbd *sdp = dip->i_sbd;
	struct qstr dotname;
	int error = 0;

	GFS_ASSERT_INODE(ip->i_di.di_entries == 2, ip,
			 gfs_dinode_print(&ip->i_di, TRUE);
	    );

	error = gfs_dir_del(dip, name);
	if (error)
		goto out;

	error = gfs_change_nlink(dip, -1);
	if (error)
		goto out;

	dotname.len = 1;
	dotname.name = ".";
	error = gfs_dir_del(ip, &dotname);
	if (error)
		goto out;

	dotname.len = 2;
	dotname.name = "..";
	error = gfs_dir_del(ip, &dotname);
	if (error)
		goto out;

	error = gfs_change_nlink(ip, -2);
	if (error)
		goto out;

	/*  This inode is being unlinked from the directory structure and
	   we need to mark that in the log so that it isn't lost during
	   a crash.  */

	gfs_trans_add_unlinked(sdp, GFS_LOG_DESC_IUL, &ip->i_num);
	set_bit(GLF_STICKY, &ip->i_gl->gl_flags);

      out:

	return error;
}

/*
 * gfs_revalidate - check to see that a inode is still in a directory
 * @dip: the directory
 * @name: the name of the file
 * @ip: the inode
 *
 * Assumes that the lock on (at least) @dip is held.
 *
 * Returns: 0 if the parent/child relationship is correct, -ENOENT if it isn't
 */

int
gfs_revalidate(struct gfs_inode *dip, struct qstr *name, struct gfs_inode *ip)
{
	struct gfs_inum inum;
	unsigned int type;
	int error;

	error = gfs_dir_search(dip, name, &inum, &type);
	if (!error) {
		if (inum.no_formal_ino == ip->i_num.no_formal_ino)
			GFS_ASSERT_INODE(ip->i_di.di_type == type, ip,);
		else
			error = -ENOENT;
	}

	return error;
}

/*
 * gfs_ok_to_move - check if it's ok to move a directory to another directory
 * @this: move this
 * @to: to here
 *
 * Follow @to back to the root and make sure we don't encounter @this
 * Assumes we already hold the rename lock.
 *
 * Returns: 0 if it's ok to move, -EXXX if it isn't
 */

int
gfs_ok_to_move(struct gfs_inode *this, struct gfs_inode *to)
{
	struct gfs_sbd *sdp = this->i_sbd;
	struct gfs_inode *tmp;
	struct gfs_holder to_gh, tmp_gh;
	struct qstr dotdot;
	int error = 0;

	memset(&dotdot, 0, sizeof (struct qstr));
	dotdot.name = "..";
	dotdot.len = 2;

	gfs_hold_istruct(to);

	for (;;) {
		if (to == this) {
			error = -EINVAL;
			break;
		}
		if (to == sdp->sd_rooti) {
			error = 0;
			break;
		}

		gfs_init_holder(to->i_gl, 0, 0, &to_gh);

		error = gfs_lookupi(&to_gh, &dotdot, NULL, &tmp_gh);
		if (error) {
			gfs_uninit_holder(&to_gh);
			break;
		}
		if (!tmp_gh.gh_gl) {
			gfs_uninit_holder(&to_gh);
			error = -ENOENT;
			break;
		}

		tmp = (struct gfs_inode *) tmp_gh.gh_gl->gl_object;

		gfs_glock_dq_uninit(&to_gh);
		gfs_glock_dq_uninit(&tmp_gh);

		gfs_put_istruct(to);
		to = tmp;
	}

	gfs_put_istruct(to);

	return error;
}

/**
 * gfs_readlinki - return the contents of a symlink
 * @ip: the symlink's inode
 * @buf: a pointer to the buffer to be filled
 * @len: a pointer to the length of @buf
 *
 * If @buf is too small, a piece of memory is gmalloc()ed and needs
 * to be freed by the caller.
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
gfs_readlinki(struct gfs_inode *ip, char **buf, unsigned int *len)
{
	struct gfs_holder i_gh;
	struct buffer_head *dibh;
	unsigned int x;
	int error;

	gfs_init_holder(ip->i_gl, LM_ST_SHARED, GL_ATIME, &i_gh);
	error = gfs_glock_nq_atime(&i_gh);
	if (error) {
		gfs_uninit_holder(&i_gh);
		goto out;
	}

	GFS_ASSERT_INODE(ip->i_di.di_size, ip,);

	error = gfs_get_inode_buffer(ip, &dibh);
	if (error)
		goto out_gunlock;

	x = ip->i_di.di_size + 1;
	if (x > *len)
		*buf = gmalloc(x);

	memcpy(*buf, (dibh)->b_data + sizeof (struct gfs_dinode), x);
	*len = x;

	gfs_drelse(dibh);

      out_gunlock:
	gfs_glock_dq_uninit(&i_gh);

      out:

	return error;
}

/**
 * gfs_glock_nq_atime - Acquire the glock and conditionally update the atime on an inode
 * @gh: the holder to acquire
 *
 * Tests atime for gfs_read, gfs_readdir and gfs_test_mmap
 * Update if the difference between the current time and the current atime 
 * is greater than a interval specfied at mount.
 *
 * Returns: 0 on success, -EXXX on error
 */

int
gfs_glock_nq_atime(struct gfs_holder *gh)
{
	struct gfs_glock *gl = gh->gh_gl;
	struct gfs_sbd *sdp = gl->gl_sbd;
	struct gfs_inode *ip;
	uint64_t curtime, quantum = sdp->sd_tune.gt_atime_quantum;
	struct buffer_head *dibh;
	unsigned int state;
	int flags;
	int was_exclusive;
	int error;

	GFS_ASSERT_GLOCK(gh->gh_flags & GL_ATIME, gl,);
	GFS_ASSERT_GLOCK(!(gh->gh_flags & GL_ASYNC), gl,);
	GFS_ASSERT_GLOCK(gl->gl_ops == &gfs_inode_glops, gl,);

	ip = (struct gfs_inode *) gl->gl_object;
	GFS_ASSERT_GLOCK(ip, gl,);

	state = gh->gh_state;
	flags = gh->gh_flags;

	error = gfs_glock_nq(gh);
	if (error)
		goto fail;

	if (test_bit(SDF_NOATIME, &sdp->sd_flags))
		goto out;

	curtime = CURRENT_TIME;
	if (curtime - ip->i_di.di_atime >= quantum) {
		was_exclusive = gl->gl_state == LM_ST_EXCLUSIVE;

		gfs_glock_dq(gh);
		gfs_reinit_holder(LM_ST_EXCLUSIVE, gh->gh_flags & ~LM_FLAG_ANY,
				  gh);
		error = gfs_glock_nq(gh);
		if (error)
			goto fail;

		/*  Verify this hasn't been updated while we were trying to get exclusive lock.  */

		curtime = CURRENT_TIME;
		if (curtime - ip->i_di.di_atime >= quantum) {
			error = gfs_trans_begin(sdp, TRANS_UPDATE_ATIME, 1, 0);
			if (error == -EROFS)
				goto out;
			if (error)
				goto fail_gunlock;

			error = gfs_get_inode_buffer(ip, &dibh);
			if (error)
				goto fail_end_trans;

			ip->i_di.di_atime = curtime;

			gfs_trans_add_bh(sdp, ip->i_gl, dibh);
			gfs_dinode_out(&ip->i_di, (dibh)->b_data);
			gfs_drelse(dibh);

			gfs_trans_end(sdp);
		}

		if (!was_exclusive) {
			gh->gh_flags |= GL_NOCACHE;
			gfs_glock_dq(gh);
			gfs_reinit_holder(state, flags, gh);
			error = gfs_glock_nq(gh);
			goto fail;	/*  Not necessarily an error  */
		}
	}

      out:

	return 0;

      fail_end_trans:
	gfs_trans_end(sdp);

      fail_gunlock:
	gfs_glock_dq(gh);

      fail:
	return error;
}

/**
 * glock_compare_atime - Compare two struct gfs_glock structures for sorting
 * @arg_a: the first structure
 * @arg_b: the second structure
 *
 */

static int
glock_compare_atime(void *arg_a, void *arg_b)
{
	struct gfs_holder *gh_a = *(struct gfs_holder **) arg_a;
	struct gfs_holder *gh_b = *(struct gfs_holder **) arg_b;
	struct lm_lockname *a = &gh_a->gh_gl->gl_name;
	struct lm_lockname *b = &gh_b->gh_gl->gl_name;
	int ret = 0;

	if (a->ln_number > b->ln_number)
		ret = 1;
	else if (a->ln_number < b->ln_number)
		ret = -1;
	else {
		/*  FixMe!!!  This is ridiculous!  */
		GFS_ASSERT_GLOCK(gh_a->gh_gl->gl_ops == &gfs_inode_glops,
				 gh_a->gh_gl,);
		GFS_ASSERT_GLOCK(gh_b->gh_gl->gl_ops == &gfs_inode_glops,
				 gh_b->gh_gl,);
		GFS_ASSERT_GLOCK(gh_a->gh_state == LM_ST_SHARED
				 || gh_a->gh_state == LM_ST_EXCLUSIVE,
				 gh_a->gh_gl,);
		GFS_ASSERT_GLOCK(gh_b->gh_state == LM_ST_SHARED
				 || gh_b->gh_state == LM_ST_EXCLUSIVE,
				 gh_b->gh_gl,);
		GFS_ASSERT_GLOCK(gh_a->gh_state == LM_ST_EXCLUSIVE
				 || !(gh_a->gh_flags & GL_LOCAL_EXCL),
				 gh_a->gh_gl,);
		GFS_ASSERT_GLOCK(gh_b->gh_state == LM_ST_EXCLUSIVE
				 || !(gh_b->gh_flags & GL_LOCAL_EXCL),
				 gh_b->gh_gl,);
		GFS_ASSERT_GLOCK(!(gh_a->gh_flags & LM_FLAG_ANY), gh_a->gh_gl,);
		GFS_ASSERT_GLOCK(!(gh_b->gh_flags & LM_FLAG_ANY), gh_b->gh_gl,);
		GFS_ASSERT_GLOCK(!((gh_a->gh_flags & GL_ATIME)
				   && (gh_b->gh_flags & GL_ATIME)),
				 gh_a->gh_gl,);

		if (gh_a->gh_state == LM_ST_SHARED
		    && gh_b->gh_state == LM_ST_EXCLUSIVE)
			ret = 1;
		else if (gh_a->gh_state == LM_ST_SHARED
			 && (gh_b->gh_flags & GL_ATIME))
			ret = 1;
	}

	return ret;
}

/**
 * gfs_glock_nq_m_atime - acquire multiple glocks where one may need an atime update
 * @num_gh: the number of structures
 * @ghs: an array of struct gfs_holder structures
 *
 * Returns: 0 on success (all glocks acquired), -EXXX on failure (no glocks acquired)
 */

int
gfs_glock_nq_m_atime(unsigned int num_gh, struct gfs_holder *ghs)
{
	struct gfs_holder *p[num_gh];
	unsigned int x;
	int error = 0;

	GFS_ASSERT(num_gh,);

	if (num_gh == 1) {
		ghs->gh_flags &= ~(LM_FLAG_TRY | GL_ASYNC);
		if (ghs->gh_flags & GL_ATIME)
			error = gfs_glock_nq_atime(ghs);
		else
			error = gfs_glock_nq(ghs);
		goto out;
	}

	for (x = 0; x < num_gh; x++)
		p[x] = &ghs[x];

	gfs_sort(p, num_gh, sizeof (struct gfs_holder *), glock_compare_atime);

	for (x = 0; x < num_gh; x++) {
		p[x]->gh_flags &= ~(LM_FLAG_TRY | GL_ASYNC);

		if (p[x]->gh_flags & GL_ATIME)
			error = gfs_glock_nq_atime(p[x]);
		else
			error = gfs_glock_nq(p[x]);

		if (error) {
			while (x--)
				gfs_glock_dq(p[x]);
			break;
		}
	}

      out:

	return error;
}

/**
 * gfs_alloc_qinode - allocate a quota inode
 * @sdp: The GFS superblock
 *
 * Returns: 0 on success, error code otherwise
 */

int
gfs_alloc_qinode(struct gfs_sbd *sdp)
{
	struct gfs_inode *dip = sdp->sd_rooti;
	struct gfs_glock *sb_gl;
	struct gfs_holder d_gh, i_gh;
	struct gfs_alloc *al;
	struct gfs_inum inum;
	struct gfs_dinode di;
	struct buffer_head *bh;
	int error;

	error = gfs_glock_nq_init(dip->i_gl, LM_ST_EXCLUSIVE, 0, &d_gh);
	if (error)
		goto fail;

	al = gfs_get_alloc(dip);

	error = gfs_hold_quota_structs(dip, 0, 0);
	if (error)
		goto fail_alloc;

	al->al_diblks = 1;
	al->al_trans = TRANS_ALLOC_QINODE;

	error = gfs_inplace_reserv(dip);
	if (error)
		goto fail_qs;

	gfs_dipick(dip, &inum);

	if (!inum.no_formal_ino)
		pick_formal_ino(sdp, &inum);

	error = gfs_glock_nq_num(sdp,
				 inum.no_formal_ino, &gfs_inode_glops,
				 LM_ST_EXCLUSIVE, GL_SKIP, &i_gh);
	if (error)
		goto fail_ipres;

	/*  Trans may require:
	   one block for the quota dinode, the superblock, RG bitmap blocks
	   for an allocation, and one block for a quota change.  */

	error =
	    gfs_trans_begin(sdp, TRANS_ALLOC_QINODE,
			    2 + al->al_rgd->rd_ri.ri_length, 1);
	if (error)
		goto fail_gunlock_gl_di;

	error = gfs_dialloc(dip, &inum, i_gh.gh_gl);
	if (error)
		goto fail_end_trans;

	error = make_dinode(dip, i_gh.gh_gl, &inum, GFS_FILE_REG, 0600, NULL);
	if (error)
		goto fail_end_trans;

	error =
	    gfs_dread(sdp, inum.no_addr, i_gh.gh_gl, DIO_START | DIO_WAIT, &bh);
	GFS_ASSERT_SBD(!error, sdp,);	/*  Already pinned  */

	gfs_dinode_in(&di, (bh)->b_data);
	di.di_flags |= GFS_DIF_JDATA;
	di.di_payload_format = GFS_FORMAT_QU;
	gfs_dinode_out(&di, (bh)->b_data);

	gfs_drelse(bh);

	error =
	    gfs_get_glstruct(sdp, GFS_SB_LOCK, &gfs_meta_glops, NO_ST_CREATE,
			     &sb_gl);
	GFS_ASSERT_SBD(!error && sb_gl, sdp,);	/*  This should already be held.  */

	error =
	    gfs_dread(sdp, GFS_SB_ADDR >> sdp->sd_fsb2bb_shift, sb_gl,
		      DIO_START | DIO_WAIT, &bh);
	if (error)
		goto fail_end_trans;

	sdp->sd_sb.sb_quota_di = inum;

	gfs_trans_add_bh(sdp, sb_gl, bh);
	gfs_sb_out(&sdp->sd_sb, (bh)->b_data);

	gfs_drelse(bh);

	gfs_put_glstruct(sb_gl);

	gfs_trans_end(sdp);

	gfs_glock_dq_uninit(&i_gh);

	gfs_inplace_relse(dip);
	gfs_unhold_quota_structs(dip);
	gfs_put_alloc(dip);

	gfs_glock_dq_uninit(&d_gh);

	gfs_log_flush(sdp);
	gfs_sync_dev(sdp);

	return 0;

      fail_end_trans:
	gfs_trans_end(sdp);

      fail_gunlock_gl_di:
	gfs_glock_dq_uninit(&i_gh);

      fail_ipres:
	gfs_inplace_relse(dip);

      fail_qs:
	gfs_unhold_quota_structs(dip);

      fail_alloc:
	gfs_put_alloc(dip);
	gfs_glock_dq_uninit(&d_gh);

      fail:
	return error;
}

/**
 * gfs_alloc_linode - allocate a license inode
 * @sdp: The GFS superblock
 *
 * Returns: 0 on success, error code otherwise
 */

int
gfs_alloc_linode(struct gfs_sbd *sdp)
{
	struct gfs_inode *dip = sdp->sd_rooti;
	struct gfs_glock *sb_gl;
	struct gfs_holder d_gh, i_gh;
	struct gfs_alloc *al;
	struct gfs_inum inum;
	struct gfs_dinode di;
	struct buffer_head *bh;
	int error;

	error = gfs_glock_nq_init(dip->i_gl, LM_ST_EXCLUSIVE, 0, &d_gh);
	if (error)
		goto fail;

	al = gfs_get_alloc(dip);

	error = gfs_hold_quota_structs(dip, 0, 0);
	if (error)
		goto fail_alloc;

	al->al_diblks = 1;
	al->al_trans = TRANS_ALLOC_LINODE;

	error = gfs_inplace_reserv(dip);
	if (error)
		goto fail_qs;

	gfs_dipick(dip, &inum);

	if (!inum.no_formal_ino)
		pick_formal_ino(sdp, &inum);

	error = gfs_glock_nq_num(sdp,
				 inum.no_formal_ino, &gfs_inode_glops,
				 LM_ST_EXCLUSIVE, GL_SKIP, &i_gh);
	if (error)
		goto fail_ipres;

	/*  Trans may require:
	   one block for the quota dinode, the superblock, RG bitmap blocks
	   for an allocation, and one block for a quota change.  */

	error =
	    gfs_trans_begin(sdp, TRANS_ALLOC_LINODE,
			    2 + al->al_rgd->rd_ri.ri_length, 1);
	if (error)
		goto fail_gunlock_gl_di;

	error = gfs_dialloc(dip, &inum, i_gh.gh_gl);
	if (error)
		goto fail_end_trans;

	error = make_dinode(dip, i_gh.gh_gl, &inum, GFS_FILE_REG, 0600, NULL);
	if (error)
		goto fail_end_trans;

	error =
	    gfs_dread(sdp, inum.no_addr, i_gh.gh_gl, DIO_START | DIO_WAIT, &bh);
	GFS_ASSERT_SBD(!error, sdp,);	/*  Already pinned  */

	gfs_dinode_in(&di, (bh)->b_data);
	di.di_flags |= GFS_DIF_JDATA;
	gfs_dinode_out(&di, (bh)->b_data);

	gfs_drelse(bh);

	error =
	    gfs_get_glstruct(sdp, GFS_SB_LOCK, &gfs_meta_glops, NO_ST_CREATE,
			     &sb_gl);
	GFS_ASSERT_SBD(!error && sb_gl, sdp,);	/*  This should already be held.  */

	error =
	    gfs_dread(sdp, GFS_SB_ADDR >> sdp->sd_fsb2bb_shift, sb_gl,
		      DIO_START | DIO_WAIT, &bh);
	if (error)
		goto fail_end_trans;

	sdp->sd_sb.sb_license_di = inum;

	gfs_trans_add_bh(sdp, sb_gl, bh);
	gfs_sb_out(&sdp->sd_sb, (bh)->b_data);

	gfs_drelse(bh);

	gfs_put_glstruct(sb_gl);

	gfs_trans_end(sdp);

	gfs_glock_dq_uninit(&i_gh);

	gfs_inplace_relse(dip);
	gfs_unhold_quota_structs(dip);
	gfs_put_alloc(dip);

	gfs_glock_dq_uninit(&d_gh);

	gfs_log_flush(sdp);
	gfs_sync_dev(sdp);

	return 0;

      fail_end_trans:
	gfs_trans_end(sdp);

      fail_gunlock_gl_di:
	gfs_glock_dq_uninit(&i_gh);

      fail_ipres:
	gfs_inplace_relse(dip);

      fail_qs:
	gfs_unhold_quota_structs(dip);

      fail_alloc:
	gfs_put_alloc(dip);
	gfs_glock_dq_uninit(&d_gh);

      fail:
	return error;
}
