/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "linux_dcache.h"
#include "dir.h"
#include "glock.h"

/**
 * gfs_drevalidate - Check directory lookup consistency
 * @dentry: the mapping to check
 * @flags: Who knows?
 *
 * Check to make sure the lookup necessary to arrive at this inode from its
 * parent is still good.
 *
 * Returns: 1 if the dentry is ok, 0 if it isn't
 */

static int
gfs_drevalidate(struct dentry *dentry, int flags)
{
	struct gfs_inode *dip;
	struct inode *inode;
	struct gfs_holder d_gh;
	struct gfs_inode *ip;
	struct gfs_inum inum;
	unsigned int type;
	int error;

	lock_kernel();

	dip = (struct gfs_inode *) dentry->d_parent->d_inode->u.generic_ip;
	GFS_ASSERT(dip,);

	if (dip->i_sbd->sd_args.ar_localcaching)
		goto valid;

	inode = dentry->d_inode;
	if (inode && is_bad_inode(inode))
		goto invalid;

	error = gfs_glock_nq_init(dip->i_gl, LM_ST_SHARED, 0, &d_gh);
	if (error)
		goto fail;

	error = gfs_dir_search(dip, &dentry->d_name, &inum, &type);
	switch (error) {
	case 0:
		if (!inode)
			goto invalid_gunlock;
		break;
	case -ENOENT:
		if (!inode)
			goto valid_gunlock;
		goto invalid_gunlock;
	default:
		goto fail_gunlock;
	}

	ip = (struct gfs_inode *) inode->u.generic_ip;
	GFS_ASSERT_SBD(ip, dip->i_sbd,);

	if (ip->i_num.no_formal_ino != inum.no_formal_ino)
		goto invalid_gunlock;

	GFS_ASSERT_INODE(ip->i_di.di_type == type, ip,);

      valid_gunlock:
	gfs_glock_dq_uninit(&d_gh);

      valid:
	unlock_kernel();
	return 1;

      invalid_gunlock:
	gfs_glock_dq_uninit(&d_gh);

      invalid:
	shrink_dcache_parent(dentry);
	if (have_submounts(dentry)) {
		unlock_kernel();
		return 1;
	}
	d_drop(dentry);

	unlock_kernel();
	return 0;

      fail_gunlock:
	gfs_glock_dq_uninit(&d_gh);

      fail:
	unlock_kernel();
	return 0;
}

struct dentry_operations gfs_dops = {
      d_revalidate:gfs_drevalidate,
};
