/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "inode.h"
#include "log.h"
#include "lops.h"
#include "unlinked.h"

/**
 * gfs_get_ulstruct - Get a structure to represent an unlinked inode
 * @sdp: the filesystem
 * @inum: the inode that's unlinked
 * @create: if TRUE, create the structure, otherwise return NULL
 *
 * Returns: the structure, or NULL
 */

struct gfs_unlinked *
gfs_get_ulstruct(struct gfs_sbd *sdp, struct gfs_inum *inum, int create)
{
	struct gfs_unlinked *ul = NULL, *new_ul = NULL;
	struct list_head *tmp, *head;

	for (;;) {
		spin_lock(&sdp->sd_unlinked_lock);

		for (head = &sdp->sd_unlinked_list, tmp = head->next;
		     tmp != head; tmp = tmp->next) {
			ul = list_entry(tmp, struct gfs_unlinked, ul_list);
			if (ul->ul_inum.no_formal_ino == inum->no_formal_ino) {
				ul->ul_count++;
				break;
			} else
				ul = NULL;
		}

		if (!ul && new_ul) {
			ul = new_ul;
			list_add(&ul->ul_list, &sdp->sd_unlinked_list);
			new_ul = NULL;
		}

		spin_unlock(&sdp->sd_unlinked_lock);

		if (ul || !create) {
			if (new_ul)
				kfree(new_ul);
			goto out;
		}

		new_ul = gmalloc(sizeof (struct gfs_unlinked));

		memset(new_ul, 0, sizeof (struct gfs_unlinked));

		new_ul->ul_count = 1;
		new_ul->ul_inum = *inum;

		INIT_LE(&new_ul->ul_new_le, &gfs_unlinked_lops);
		INIT_LE(&new_ul->ul_incore_le, &gfs_unlinked_lops);
		INIT_LE(&new_ul->ul_ondisk_le, &gfs_unlinked_lops);
	}

      out:

	return ul;
}

/**
 * gfs_hold_ulstruct - increment the usage count on a struct gfs_unlinked
 * @sdp: the filesystem
 * @ul: the structure
 *
 */

void
gfs_hold_ulstruct(struct gfs_sbd *sdp, struct gfs_unlinked *ul)
{

	spin_lock(&sdp->sd_unlinked_lock);
	ul->ul_count++;
	spin_unlock(&sdp->sd_unlinked_lock);

}

/**
 * gfs_put_ulstruct - decrement the usage count on a struct gfs_unlinked
 * @sdp: the filesystem
 * @ul: the structure
 *
 * Free the structure if its reference count hits zero.
 *
 */

void
gfs_put_ulstruct(struct gfs_sbd *sdp, struct gfs_unlinked *ul)
{

	spin_lock(&sdp->sd_unlinked_lock);

	GFS_ASSERT_SBD(ul->ul_count, sdp,);
	ul->ul_count--;

	if (!ul->ul_count) {
		GFS_ASSERT_SBD(!test_bit(ULF_IC_LIST, &ul->ul_flags) &&
			       !test_bit(ULF_OD_LIST, &ul->ul_flags) &&
			       !test_bit(ULF_LOCK, &ul->ul_flags), sdp,);
		list_del(&ul->ul_list);
		spin_unlock(&sdp->sd_unlinked_lock);
		kfree(ul);
	} else
		spin_unlock(&sdp->sd_unlinked_lock);

}

/**
 * gfs_find_ulstruct - Find a inode to try to deallocate
 * @sdp: the filesystem
 *
 * The returned structure is locked and needs to be unlocked
 * with gfs_unlock_ulstruct().
 *
 * Returns: A unlinked structure, or NULL
 */

struct gfs_unlinked *
gfs_find_ulstruct(struct gfs_sbd *sdp)
{
	struct list_head *tmp, *head;
	struct gfs_unlinked *ul = NULL;

	if (test_bit(SDF_ROFS, &sdp->sd_flags))
		goto out;

	gfs_log_lock(sdp);
	spin_lock(&sdp->sd_unlinked_lock);

	if (!atomic_read(&sdp->sd_unlinked_ic_count))
		goto out_unlock;

	for (head = &sdp->sd_unlinked_list, tmp = head->next; tmp != head;
	     tmp = tmp->next) {
		ul = list_entry(tmp, struct gfs_unlinked, ul_list);

		if (test_bit(ULF_LOCK, &ul->ul_flags))
			continue;
		if (!test_bit(ULF_IC_LIST, &ul->ul_flags))
			continue;

		list_del(&ul->ul_list);
		list_add_tail(&ul->ul_list, &sdp->sd_unlinked_list);

		set_bit(ULF_LOCK, &ul->ul_flags);
		ul->ul_count++;

		goto out_unlock;
	}

	ul = NULL;

      out_unlock:
	spin_unlock(&sdp->sd_unlinked_lock);
	gfs_log_unlock(sdp);

      out:

	return ul;
}

/**
 * gfs_unlock_ulstruct - drop and a reference on a unlinked structure
 * @sdp: the filesystem
 * @ul: the unlinked inode structure
 *
 */

void
gfs_unlock_ulstruct(struct gfs_sbd *sdp, struct gfs_unlinked *ul)
{
	spin_lock(&sdp->sd_unlinked_lock);

	GFS_ASSERT_SBD(test_bit(ULF_LOCK, &ul->ul_flags), sdp,);
	clear_bit(ULF_LOCK, &ul->ul_flags);

	GFS_ASSERT_SBD(ul->ul_count, sdp,);
	ul->ul_count--;

	if (!ul->ul_count) {
		GFS_ASSERT_SBD(!test_bit(ULF_IC_LIST, &ul->ul_flags) &&
			       !test_bit(ULF_OD_LIST, &ul->ul_flags), sdp,);
		list_del(&ul->ul_list);
		spin_unlock(&sdp->sd_unlinked_lock);
		kfree(ul);
	} else
		spin_unlock(&sdp->sd_unlinked_lock);
}

/**
 * gfs_unlinked_merge - add/remove a unlinked inode from the in-memory list
 * @sdp: the filesystem
 * @type: is this a unlink tag or a dealloc tag
 * @inum: the inode number
 *
 */

void
gfs_unlinked_merge(struct gfs_sbd *sdp, unsigned int type,
		   struct gfs_inum *inum)
{
	struct gfs_unlinked *ul;

	GFS_ASSERT_SBD(atomic_read(&sdp->sd_unlinked_ic_count) ==
		       atomic_read(&sdp->sd_unlinked_od_count), sdp,);

	ul = gfs_get_ulstruct(sdp, inum, ST_CREATE);

	gfs_log_lock(sdp);

	switch (type) {
	case GFS_LOG_DESC_IUL:
		gfs_hold_ulstruct(sdp, ul);
		gfs_hold_ulstruct(sdp, ul);
		GFS_ASSERT_SBD(!test_bit(ULF_IC_LIST, &ul->ul_flags) &&
			       !test_bit(ULF_OD_LIST, &ul->ul_flags), sdp,);
		set_bit(ULF_IC_LIST, &ul->ul_flags);
		set_bit(ULF_OD_LIST, &ul->ul_flags);
		atomic_inc(&sdp->sd_unlinked_ic_count);
		atomic_inc(&sdp->sd_unlinked_od_count);

		break;

	case GFS_LOG_DESC_IDA:
		GFS_ASSERT_SBD(test_bit(ULF_IC_LIST, &ul->ul_flags) &&
			       test_bit(ULF_OD_LIST, &ul->ul_flags), sdp,);
		clear_bit(ULF_IC_LIST, &ul->ul_flags);
		clear_bit(ULF_OD_LIST, &ul->ul_flags);
		gfs_put_ulstruct(sdp, ul);
		gfs_put_ulstruct(sdp, ul);
		GFS_ASSERT_SBD(atomic_read(&sdp->sd_unlinked_ic_count), sdp,);
		atomic_dec(&sdp->sd_unlinked_ic_count);
		GFS_ASSERT_SBD(atomic_read(&sdp->sd_unlinked_od_count), sdp,);
		atomic_dec(&sdp->sd_unlinked_od_count);

		break;
	}

	gfs_log_unlock(sdp);

	gfs_put_ulstruct(sdp, ul);

}

/**
 * gfs_unlinked_cleanup - get rid of any extra struct gfs_unlinked structures
 * @sdp: the filesystem
 *
 */

void
gfs_unlinked_cleanup(struct gfs_sbd *sdp)
{
	struct gfs_unlinked *ul;

      restart:
	gfs_log_lock(sdp);

	GFS_ASSERT_SBD(atomic_read(&sdp->sd_unlinked_ic_count) ==
		       atomic_read(&sdp->sd_unlinked_od_count), sdp,);

	spin_lock(&sdp->sd_unlinked_lock);

	while (!list_empty(&sdp->sd_unlinked_list)) {
		ul = list_entry(sdp->sd_unlinked_list.next, struct gfs_unlinked,
				ul_list);

		if (ul->ul_count > 2) {
			spin_unlock(&sdp->sd_unlinked_lock);
			gfs_log_unlock(sdp);
			current->state = TASK_UNINTERRUPTIBLE;
			schedule_timeout(HZ);
			goto restart;
		}
		GFS_ASSERT_SBD(ul->ul_count == 2, sdp,);

		GFS_ASSERT_SBD(test_bit(ULF_IC_LIST, &ul->ul_flags) &&
			       test_bit(ULF_OD_LIST, &ul->ul_flags) &&
			       !test_bit(ULF_LOCK, &ul->ul_flags), sdp,);

		list_del(&ul->ul_list);

		atomic_dec(&sdp->sd_unlinked_ic_count);
		atomic_dec(&sdp->sd_unlinked_od_count);

		spin_unlock(&sdp->sd_unlinked_lock);
		kfree(ul);
		spin_lock(&sdp->sd_unlinked_lock);
	}

	spin_unlock(&sdp->sd_unlinked_lock);

	GFS_ASSERT_SBD(!atomic_read(&sdp->sd_unlinked_ic_count) &&
		       !atomic_read(&sdp->sd_unlinked_od_count), sdp,);

	gfs_log_unlock(sdp);

}

/**
 * gfs_unlinked_dealloc - Go through the list of inodes to be deallocated
 * @sdp: the filesystem
 *
 * Returns: 0 on success, -EXXX on failure
 */

void
gfs_unlinked_dealloc(struct gfs_sbd *sdp)
{
	unsigned int hits, strikes;
	int error;

	for (;;) {
		hits = 0;
		strikes = 0;

		for (;;) {
			error = gfs_try_dealloc_inode(sdp);
			if (!error) {
				hits++;
				if (strikes)
					strikes--;
			} else if (error == 1) {
				strikes++;
				if (strikes >=
				    atomic_read(&sdp->sd_unlinked_ic_count)) {
					error = 0;
					break;
				}
			} else if (error == 2) {
				error = 0;
				goto out;
			} else
				goto out;
		}

		if (!hits || !test_bit(SDF_INODED_RUN, &sdp->sd_flags))
			break;

		yield();
	}

      out:
	if (error && error != -EROFS)
		printk("GFS: fsid=%s: error deallocating inodes: %d\n",
		       sdp->sd_fsname, error);

}
