/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

/* libpool.h - exported header file for libpool.  */

#ifndef __LIBPOOL_H
#define __LIBPOOL_H

#include "osi_user.h"
#include <string.h>
#include "pool_std.h"

/* FIXME: do i want to redefine this in userspace? */
#define POOL_NAME_SIZE (256)

#define POOL_DEBUG   1
#define POOL_QUIET   2
#define POOL_VERBOSE 4
#define POOL_FORCE   8

#define POOL_LINE_MAX 256

/* Return values from library calls */
#define STD_ERROR            1
#define IOCTL_FAILED         2
#define COMM_DEV_NOT_OPEN    3

#define UNKNOWN_LABEL "unknown"
#define ERROR_LABEL "error"

/* This stuff should go away */
#if 1
#define debug_print(fmt, args...) \
do \
{ \
  if(check_flag(POOL_DEBUG)) { \
    fprintf(stderr, "%s:%d - ", __FILE__, __LINE__); \
    fprintf(stderr, fmt, ##args); \
  } \
} \
while(0)

#define std_print(fmt, args...) \
do \
{ \
  if(!(check_flag(POOL_QUIET))) { \
    fprintf(stdout, fmt, ##args); \
  } \
} \
while(0)

#define verbose_print(fmt, args...) \
do \
{ \
  if((check_flag(POOL_VERBOSE))) { \
    fprintf(stdout, fmt, ##args); \
  } \
} \
while(0)

#define err_print(fmt, args...) \
do \
{ \
   fprintf(stderr, fmt, ##args); \
} \
while(0)

#endif /* 1 - change to 0 when new logging goes in */

typedef struct diskinfo diskinfo_t;
typedef struct disklabels disklabels_t;
typedef struct user_pool user_pool_t;
typedef struct user_subpool user_subpool_t;
typedef struct user_device user_device_t;
typedef struct user_poolstat user_poolstat_t;

enum label_type {
        unknown_label,
        error_label,
        known_label,
        pool_label
};

struct diskinfo {
   int          dl_count;
   int          pl_count;
   struct disklabels *dl;
   struct user_pool  *pl;
};

struct disklabels {
   char *name;
   char *dev_path;
   enum label_type lt;
};

struct user_pool {
   uint64 id;
   char   name[POOL_NAME_SIZE];
   uint32 major;
   uint32 minor;
   uint32 num_sps;
   uint32 in_use;
   uint32 mptype;
   uint32 mpstripe;
   user_subpool_t *sp;
};

struct user_subpool {
   uint32 id;
   uint32 striping;
   uint32 num_devs;
   uint32 type;
   user_device_t *devs;
};

struct user_device {
   uint32 initialized;
   uint32 sp_id;
   uint32 devid;
   uint32 path_count;
   uint32 good_path_count;
   uint32 dummy;
   uint64 blocks;
   char *dev_path[POOL_MAX_MULTIPATH];
};

struct user_poolstat {
   uint32 reads;
   uint32 readblks;
   uint32 writes;
   uint32 writeblks;
};

/* Misc fxns */
inline int check_flag(int flag);

inline int add_flag(int flag);

inline int del_flag(int flag);

int get_sptype_num(char *type, uint32 *num);

int get_sptype_name(uint32 type, char **name);

void destroy_sptype_name(char **name);

void destroy_di(diskinfo_t *di);

/* Kernel related fxns */

int open_comm(struct sstring *emsg);

int close_comm(struct sstring *emsg);

int check_version(struct sstring *emsg);

int reintegrate_paths(char *name, struct sstring *emsg);

int tune_multipath(char *name, int mp, struct sstring *emsg);

int assemble_pool(char *name, struct sstring *emsg);

int grow_pool(char *name, struct sstring *emsg);

int disassemble_pool(char *name, struct sstring *emsg);

int clear_pool_stats(char *name, struct sstring *emsg);

int get_pool_stats(char *name, user_poolstat_t **stat, struct sstring *emsg);

void destroy_pool_stats(user_poolstat_t **stat);

int update_device_nodes(struct sstring *emsg);

int get_loaded_pools(diskinfo_t *di, struct sstring *emsg);

int get_pool_alias(char **pool, uint32 minor, struct sstring *emsg);

void destroy_pool_alias(char *pool);

int clean_pool_dir(struct sstring *emsg);

/* Pool config related fxns */
int read_pool_config(char *cfg_file, user_pool_t **p);
int print_config(user_pool_t *pool);

/* Label related fxns */
int print_pool_labels(char *name, struct sstring *emsg);

int scan_disks(diskinfo_t *di, struct sstring *emsg);

int write_labels(user_pool_t *p, struct sstring *emsg);

int erase_label(user_pool_t *p, struct sstring *emsg);

int rename_pool(char *oldpool, char *name, struct sstring *emsg);

int reminor_pool(char *pool, uint16 num, struct sstring *emsg);

#endif /* __LIBPOOL_H */

/*
 * vim: et ts=3 sw=3 noai cin :
 */

