/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

/* pool_tune.c - Tune in-kernel pool information */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <ctype.h>

#include "sisstring.h"
#include "sislogger.h"
#include "siscli.h"
#include "libpool.h"
#include "osi_user.h"
#include "pool_release.cf"
#include "copyright.cf"

#define STRING_SIZE 512

void _usage(char *name) {

   printf("Usage:\n");
   printf("\n");
   printf("%s [options] [pool(s)]\n", name);
   printf("\n");
   printf("Options:\n");
   printf("  -D               Enable debugging output.\n");
   printf("  -h               This usage information.\n");
   printf("  -m <name/num>    Tune multipathing -\n"
          "                   <name>: \"none\" or \"failover\"\n"
          "                   <num>: stripe size(kB) for round-robin"
          " multipathing\n");
   printf("  -O               Override prompts.\n");
   printf("  -q               Enable quiet output.\n");
   printf("  -r               Restore failed paths.\n");
   printf("  -V               Version information.\n");
   printf("  -v               Enable verbose output.\n");

   return;
}

/* don't expect the value returned by this function to hang on for long...*/
char *_get_mp(int mp) {

   if (mp < 0)
      return "none";
   else if (mp == 0)
      return "failover";
   else
      return "round-robin";
}

int main(int argc, char **argv)
{
   int    error = 0, tmperr;
   int    c;
   int    quiet = 0;
   int    verbose = 0;
   int    i = 0;
   int    j = 0;
   int    len = 0;
   int    mp = 0;
   char   action = 0,
        **arg_list = NULL;
   char   msg[STRING_SIZE];
   diskinfo_t di;
   struct sstring *emsg = NULL;

   while((c = getopt(argc, argv, "Dhm:OqrVv")) != -1 && !error) {
      switch(c) {

      case 'D':
         verbose = 4;
         break;

      case 'h':
         _usage(argv[0]);
         return error;
         break;

      case 'm':
         if(action != 0) {
            mutually_exclusive(argv[0], c, action);
            error = 1;
         }
         else {
            action = c;
            if(strcmp(optarg, "none") == 0) {
               mp = -1;
            }
            else if (strcmp(optarg, "failover") == 0) {
               mp = 0;
            }
            else {
               mp = atoi(optarg);
            }
         }
         break;

      case 'O':
         add_flag(POOL_FORCE);
         break;

      case 'q':
         if(verbose > 0) {
            mutually_exclusive(argv[0], c, 'v');
            error = 1;
         }
         else {
            quiet++;
         }
         break;

      case 'r':
         if(action != 0) {
            mutually_exclusive(argv[0], c, action);
            error = 1;
         }
         else {
            action = c;
         }
         break;

      case 'V':
         printf("%s %s (built %s %s)\n", argv[0],
                POOL_RELEASE_NAME, __DATE__, __TIME__);
         printf("%s\n", GFS_COPYRIGHT);
         return error;
         break;

      case 'v':
         if(quiet > 0) {
            mutually_exclusive(argv[0], c, 'q');
            error = 1;
         }
         else {
            verbose++;
            break;
         }

      case ':':
      case '?':
         fprintf(stderr, "Please use '-h' for usage.\n");
         exit(1);
         break;

      default:
         fprintf(stderr, "Bad programmer! You forgot to catch the"
                 " %c flag\n", c);
         exit(1);
         break;
      }


   }

   if(argc > optind) {
      arg_list = &(argv[optind]);
      argc = argc - optind;
   }

   if(!error) {
      if((error = ccs_initlog("pool.ccs", "pool/tool_config",
                              verbose, quiet))) {
         if((error = init_defaultlog(verbose, quiet, NULL)) != 0) {
            goto end;
         }
      }

      emsg = ss_init(STRING_SIZE-1);

      if((error = open_comm(emsg))) {

         if(error == ENXIO) {
            log_err("Unable to open pool comm device.  "
                    "Verify pool module is loaded");
            cstring2ss(emsg, "", 0);
         }
         else {
            ss_len(emsg, &len);
            if(len > 0) {
               ss2cstring(msg, STRING_SIZE, emsg);
               log_crit(msg);
            }
         }
         goto end;
      }

      if ((error = check_version(emsg))) {
         goto end;
      }


      switch(action) {

         /* Change multipathing  */
      case 'm':
         if((error = get_loaded_pools(&di, emsg)) == 0) {
            if(arg_list != NULL) {
               for(i = 0; i < argc; i++) {

                  for(j = 0; j < di.pl_count; j++) {
                     if(strcmp(arg_list[i], di.pl[j].name) == 0) {
                        break;
                     }
                  }

                  if(i == di.pl_count) {
                     log_err("No pool named %s found", arg_list[i]);
                     error |= 1;
                     continue;
                  }

                  if((tmperr = tune_multipath(arg_list[i], mp, emsg)) != 0) {
                     error |= tmperr;
                     ss_len(emsg, &len);
                     if(len > 0) {
                        ss2cstring(msg, STRING_SIZE, emsg);
                        log_err(msg);
                     }
                     else {
                        log_err("Unable to adjust multipathing for %s",
                                arg_list[i]);
                     }
                  }
                  else {
                     log_warn("Multipathing for %s changed to '%s'", arg_list[i],
                              _get_mp(mp));
                  }
               }
            }
            else {
               for(i = 0; i < di.pl_count; i++) {
                  if((tmperr= tune_multipath(di.pl[i].name, mp, emsg)) != 0){
                     error |= tmperr;
                     ss_len(emsg, &len);
                     if(len > 0) {
                        ss2cstring(msg, STRING_SIZE, emsg);
                        log_err(msg);
                     }
                     else {
                        log_err("Unable to adjust multipathing for %s",
                                arg_list[i]);
                     }
                  }
                  else {
                     log_warn("Multipathing for %s changed to '%s'",
                              di.pl[i].name, _get_mp(mp));
                  }
               }
            }
         }
         break;

         /* Reintegrate paths */
      case 'r':
         if((error = get_loaded_pools(&di, emsg)) == 0) {
            if(arg_list != NULL) {
               for(i = 0; i < argc; i++) {
                  if((tmperr = reintegrate_paths(arg_list[i], emsg)) != 0) {
                     error |= tmperr;
                     ss_len(emsg, &len);
                     if(len > 0) {
                        ss2cstring(msg, STRING_SIZE, emsg);
                        log_err(msg);
                     }
                     else {
                        log_err("Unable to reintegrate paths for %s",
                                arg_list[i]);
                     }
                  }
                  else {
                     log_warn("Successfully reintegrated paths for %s",
                              arg_list[i]);
                  }
               }
            }
            else {
               for(i = 0; i < di.pl_count; i++) {
                  if((tmperr = reintegrate_paths(di.pl[i].name, emsg)) != 0) {
                     error |= tmperr;
                     ss_len(emsg, &len);
                     if(len > 0) {
                        ss2cstring(msg, STRING_SIZE, emsg);
                        log_err(msg);
                     }
                     else {
                        log_err("Unable to reintegrate paths for %s",
                                di.pl[i].name);
                     }
                  }
                  else {
                     log_warn("Successfully reintegrated paths for %s",
                              di.pl[i].name);
                  }
               }
            }
         }
         break;

      default:
         log_err("Please use '-h' for usage.");
         error = 1;
         break;
      }
   }
 end:
   close_comm(emsg);

   /* Until we have well defined error codes, the exit status will either
      be 0 or 1 */
   if(error != 0) {
      error = 1;
   }
   return error;
}
