/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**  
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <wait.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <linux/raw.h>
#include <linux/kdev_t.h>

#include "ccs_ft.h"

#include "copyright.cf"
#include "ccs_release.cf"
#include "ccs_log.h"
#include "ccs_ondisk.h"
#include "ccs_lockfile.h"
#include "ccs_iov.h"

#define DEFAULT_BBS 16384
#define CCSD_LOCKFILE "/var/run/sistina/ccsd.pid"

#ifndef PAGE_SIZE
#define PAGE_SIZE (65536)
#endif

static int be_port = CCS_DEFAULT_BEPORT;
static int fe_port = CCS_DEFAULT_FEPORT;
static char *cluster_name = NULL;

/* only one of the following access methods should be used */
static char *local = NULL;
static char *server = NULL;
static char *device = NULL;
static int device_fd=-1;  /* hold device open so modules can be removed */
static int verbose=0;

static void _destroy_lockfile(void);
static void print_start_msg(void);
static void parent_exit_handler(int sig);
static int set_signals(void);
static void print_usage(FILE *stream);
static int bind_to_raw_device(char *dev, char *rdev);
static void unbind_raw_device(char *dev);
static int is_valid_local(char *local);
static int is_valid_device(char *device);
static int is_valid_server(char *server);
static int process_local_request(int afd);
static int process_server_request(int afd);
static int process_device_request(int afd);


int main(int argc, char *argv[]){
  int c;
  int daemon = 1;

  struct sockaddr_in addr;
  int fe_sfd, fe_afd;
  int parent_pid=getpid();

  while((c = getopt(argc, argv, "c:d:f:hlns:vV")) != -1){
    switch(c){
    case 'c':
      cluster_name = optarg;
      break;

    case 'd':
      if(local || server){
	fprintf(stderr, "-d, -f, and -s are mutually exclusive.\n"
		"Try 'ccsd -h' for more info.\n");
	exit(EXIT_FAILURE);
      } else {
	device = optarg;
      }
      break;

    case 'l':
      cluster_name = "__sentinel__";
      break;

    case 'h':
      print_usage(stdout);
      exit(EXIT_SUCCESS);

    case 'f':
      if(server || device){ 
	fprintf(stderr, "-d, -f, and -s are mutually exclusive.\n"
		"Try 'ccsd -h' for more info.\n");
	exit(EXIT_FAILURE);
      } else {
	local = optarg;
      }
      break;

    case 'n':
      daemon = 0;
      break;

    case 's':
      if(local || device){
	fprintf(stderr, "-d, -f, and -s are mutually exclusive.\n"
		"Try 'ccsd -h' for more info.\n");
	exit(EXIT_FAILURE);
      } else {
	char *tmp;
	server = optarg;
	tmp = strstr(server, ":");
	if(tmp){
	  tmp[0] = '\0';
	  tmp++;
	  be_port = atoi(tmp);
	}
      }
      break;
    case 'v':
      verbose=1;
      break;
    case 'V':
      printf("%s %s (built %s %s)\n", argv[0], CCS_RELEASE_NAME,
	     __DATE__, __TIME__);
      printf("%s\n", SISTINA_COPYRIGHT);
      exit(EXIT_SUCCESS);
      break;
    default:
      print_usage(stderr);
      exit(EXIT_FAILURE);
    }
  }

  if(!server && !device && !local && !cluster_name){
      fprintf(stderr, "One of the following options must be specified: "
	      "-d, -f, -s\n" 
	      "Try 'ccsd -h' for more info.\n");
    exit(EXIT_FAILURE);
  }

  if(!device && !local && !cluster_name){
    fprintf(stderr, "The cluster name must be specified.\n"
	    "Try 'ccsd -h' for more info.\n");
    exit(EXIT_FAILURE);
  }

  if(argc > optind){
    fprintf(stderr, "Too many arguments.\n"
	    "Try 'ccsd -h' for help.\n");
    exit(EXIT_FAILURE);
  }    

  if(local){
    if(local[0] != '/'){
      char tmp[256];
      if(!getcwd(tmp, 256)){
	fprintf(stderr, "Unable to get current working directory.\n");
	exit(EXIT_FAILURE);
      }
      sprintf(tmp+strlen(tmp), "/%s", local);
      local = strdup(tmp);
      if(!local){
	fprintf(stderr, "Unable to allocate memory for archive path.\n");
	exit(EXIT_FAILURE);
      }
    }

    if(!is_valid_local(local)){
      exit(EXIT_FAILURE);
    }
  } else if(device){
    char raw_device[32];
    if(device[0] != '/'){
      char tmp[256];
      if(!getcwd(tmp, 256)){
	fprintf(stderr, "Unable to get current working directory.\n");
	exit(EXIT_FAILURE);
      }
      sprintf(tmp+strlen(tmp), "/%s", device);
      device = strdup(tmp);
      if(!device){
	fprintf(stderr, "Unable to allocate memory for archive path.\n");
	exit(EXIT_FAILURE);
      }
    }

    if(!is_valid_device(device)){
      exit(EXIT_FAILURE);
    }
    
    if(bind_to_raw_device(device, raw_device)){
      exit(EXIT_FAILURE);
    }
    device_fd = open(raw_device, O_RDONLY/* | O_DIRECT*/);
    if(device_fd < 0){
      fprintf(stderr, "Unable to open %s (%s): %s\n",
	      device, raw_device, strerror(errno));
      exit(EXIT_FAILURE);
    }
  } else if(server){
    if(!is_valid_server(server)){
      exit(EXIT_FAILURE);
    }
  } else {
    if(cluster_name){
      if(!strcmp(cluster_name, "__sentinel__")){
	fprintf(stderr,
		"The '-l' option requires the use of the '-s' option.\n");
      } else {
	fprintf(stderr,
		"The '-c' option requires the use of the '-s' option.\n");
      }
    } else {
      fprintf(stderr, "One of the following options must be specified: "
	      "-d, -f, -s\n" 
	      "Try 'ccsd -h' for more info.\n");
    }
    exit(EXIT_FAILURE);
  }    

  if(cluster_name && !strcmp(cluster_name, "__sentinel__")){
    exit(EXIT_SUCCESS);
  }

  openlog("ccsd", LOG_PID, LOG_DAEMON);

  if(daemon){
    int pid;

    if(test_lockfile(CCSD_LOCKFILE)){
      exit(EXIT_FAILURE);
    }

    signal(SIGTERM, &parent_exit_handler);

    if((pid = fork()) < 0){
      fprintf(stderr, "Unable to fork().\n");
      exit(EXIT_FAILURE);
    } else if (pid !=0){
      while(1){
	sleep(30);
      }
      /* close the parent */
      exit(EXIT_SUCCESS);
    }

    setsid();
    chdir("/");
    umask(0);

    close(0); close(1); close(2);
    open("/dev/null", O_RDONLY);  /* reopen stdin */
    open("/dev/null", O_WRONLY);  /* reopen stdout */
    open("/dev/null", O_WRONLY);  /* reopen stderr */    

    if(create_lockfile(CCSD_LOCKFILE)){
      exit(EXIT_FAILURE);
    }
  }

  atexit(&_destroy_lockfile);  /* destroy lockfile at exit */

  set_signals();
  print_start_msg();

  /* open the Front End socket */
  fe_sfd = socket(PF_INET, SOCK_STREAM, 0);

  if(fe_sfd < 0){
    /* if daemon, report to syslog */
    log_sys_err("Socket creation failed");
    exit(EXIT_FAILURE);
  } else {
    int trueint = 1;
    setsockopt(fe_sfd, SOL_SOCKET, SO_REUSEADDR, &trueint, sizeof(int));
  }

  addr.sin_family = AF_INET;
  inet_aton("127.0.0.1", (struct in_addr *)&addr.sin_addr.s_addr);
  //  addr.sin_addr.s_addr = INADDR_ANY;
  addr.sin_port = htons(fe_port);

  if(bind(fe_sfd, (struct sockaddr *)&addr, sizeof(struct sockaddr_in)) < 0){
    log_sys_err("Unable to bind socket");
    close(fe_sfd);
    exit(EXIT_FAILURE);
  }

  listen(fe_sfd, 5);

  if(daemon) kill(parent_pid, SIGTERM);

  while(1){
    int pid;
    int len = sizeof(struct sockaddr_in);
    
    fe_afd = accept(fe_sfd, (struct sockaddr *)&addr, &len);

    if(fe_afd < 0){
      log_sys_err("Unable to accept connection");
      continue;
    }

    if(ntohs(addr.sin_port) > 1024){
      log_err("Refusing connection from port > 1024:  port = %d",
	      ntohs(addr.sin_port));
      close(fe_afd);
      continue;
    }

    pid = fork();

    if(pid < 0){
      log_sys_err("Unable to fork");
      /* Rely on exit to close sfd and afd */
      exit(EXIT_FAILURE);
    } else if(pid){
      /* parent doesn't need afd */
      close(fe_afd);
    } else {
      /* child doesn't need sfd */
      close(fe_sfd);
      break;
    }
    wait(NULL);
  }

  if(verbose){
    unsigned char *saddr = (char *)(&addr.sin_addr);
    log_msg("Connection accepted from host: %u.%u.%u.%u.\n",
	    saddr[0],saddr[1],saddr[2],saddr[3]);
  }

  if(local){
    if(process_local_request(fe_afd)){
      close(fe_afd);
      exit(EXIT_FAILURE);
    }
  } else if(device){
    if(process_device_request(fe_afd)){
      close(fe_afd);
      exit(EXIT_FAILURE);
    }
  } else {
    if(process_server_request(fe_afd)){
      close(fe_afd);
      exit(EXIT_FAILURE);
    }
  }

  if(device_fd >= 0){
    close(device_fd);
  }

  close(fe_afd);

  exit(EXIT_SUCCESS);
}

static void _destroy_lockfile(void){
  destroy_lockfile(CCSD_LOCKFILE);
}

static void print_start_msg(void){
  log_msg("Starting ccsd %s:\n", CCS_RELEASE_NAME);
  log_msg(" Built: " __DATE__" " __TIME__"\n");
  log_msg(" %s\n", SISTINA_COPYRIGHT);
}

static void parent_exit_handler(int sig){
  exit(EXIT_SUCCESS);
}

static void sig_handler(int sig){
  int err;
  switch(sig){
  case SIGINT:
    log_msg("Stopping ccsd, SIGINT received.\n");
    err = EXIT_SUCCESS;
    break;
  case SIGQUIT:
    log_msg("Stopping ccsd, SIGQUIT received.\n");
    err = EXIT_SUCCESS;
    break;
  case SIGTERM:
    log_msg("Stopping ccsd, SIGTERM received.\n");
    err = EXIT_SUCCESS;
    break;
  case SIGSEGV:
    log_err("Stopping ccsd, SIGSEGV received.\n");
    err = EXIT_FAILURE;
    break;
  default:
    log_err("Stopping ccsd, unknown signal received.\n");
    err = EXIT_FAILURE;
  }
  if(device_fd >= 0 && device){
    close(device_fd);
  }
  exit(err);
}

static int set_signals(void){
  signal(SIGINT, &sig_handler);
  signal(SIGQUIT, &sig_handler);
  signal(SIGTERM, &sig_handler);
  signal(SIGSEGV, &sig_handler);
  return 0;
}  

static void print_usage(FILE *stream){
  fprintf(stream, 
  "Usage:\n\n"
  "ccsd -c <cluster> <-s <server>[:port]> [hlnVv]\n"
  "ccsd -d <device> [hnVv]\n"
  "ccsd -f <file> [hnVv]\n\n"
  "Options:\n"
  " -c <cluster>    Specify the cluster name.\n"
  " -d <device>     Device on which the configuration archive resides.\n"
  " -f <file>       Local file containing the configuration archive.\n"
  " -h              Print this help, then exit.\n"
  " -l              List available clusters on CCS server.\n"
  " -n              No Daemon.  Run in the foreground.\n"
  " -s <IP:Port>    Server on which the configuration archive resides.\n"
  " -V              Print version information.\n"
  " -v              Verbose.\n"
  );
}


/* unbind_raw_device
 * @dev: the device to unbind
 *
 * This function takes the name of the device that is bound to a
 * raw device (e.g. /dev/pool/brassow) and releases the raw
 * device from use.
 */
static void unbind_raw_device(char *dev){
  int raw_fd, i;
  int new_raw_method=1;
  int major, minor;
  struct raw_config_request raw_req;
  struct stat stat_buf;

  if(stat(dev, &stat_buf)){
    fprintf(stderr, "Unable to stat %s: %s\n",
	    dev, strerror(errno));
    return;
  }
  if(!S_ISBLK(stat_buf.st_mode)){
    fprintf(stderr, "%s is not a block device.\n", dev);
    return;
  } else {
    major = major(stat_buf.st_rdev);
    minor = minor(stat_buf.st_rdev);
  }

  /* Are we using the old way, or the new way? */
  if(stat("/dev/raw", &stat_buf)){
    fprintf(stderr, "Unable to stat /dev/raw: %s\n",
	    strerror(errno));
    return;
  }
  if(S_ISDIR(stat_buf.st_mode)){  /* The new way */
    new_raw_method = 1;
    if((raw_fd = open("/dev/rawctl", O_RDWR)) < 0){
      fprintf(stderr, "Unable to open raw control device: %s\n",
	      strerror(errno));
      return;
    }
  } else {  /* The old way */
    new_raw_method = 0;
    if((raw_fd = open("/dev/raw", O_RDWR)) < 0){
      fprintf(stderr, "Unable to open raw control device: %s\n",
	      strerror(errno));
      return;
    }
  }

  for(i=64; i<256; i++){
    raw_req.raw_minor = i;
    if(ioctl(raw_fd, RAW_GETBIND, &raw_req)< 0){
      fprintf(stderr, "Unable to perform ioctl on raw control device: %s\n",
	      strerror(errno));
      close(raw_fd);
      return;
    }
    if(raw_req.block_major == major && raw_req.block_minor == minor){
      raw_req.block_major = raw_req.block_minor = 0;
      if(ioctl(raw_fd, RAW_SETBIND, &raw_req)< 0){
	if(errno != EBUSY){
	  fprintf(stderr,"Unable to perform ioctl on raw control device: %s\n",
		  strerror(errno));
	}
	close(raw_fd);
	return;
      }
    }
  }
  close(raw_fd);
}


static void _unbind_raw_device(int unused, void *arg){
  unbind_raw_device((char *)arg);
}


/* bind_to_raw_device
 * @dev: the block device to bind to raw device
 * @rdev: the raw device assigned - space must be allocated
 *
 * This function binds a block device to a specific raw device.  The
 * name of the raw device is returned in the space provided.
 *
 * Returns: 0 on success, -EXXX on error
 */
static int bind_to_raw_device(char *dev, char *rdev){
  int raw_fd, i, found=0;
  int new_raw_method=1;
  int major, minor;
  struct raw_config_request raw_req;
  struct stat stat_buf;

  if(stat(dev, &stat_buf)){
    fprintf(stderr, "Unable to stat %s: %s\n",
	    dev, strerror(errno));
    return -errno;
  }
  if(!S_ISBLK(stat_buf.st_mode)){
    fprintf(stderr, "%s is not a block device.\n", dev);
    return -ENOTBLK;
  } else {
    major = major(stat_buf.st_rdev);
    minor = minor(stat_buf.st_rdev);
  }

  /* Are we using the old way, or the new way? */
  if(stat("/dev/raw", &stat_buf)){
    fprintf(stderr, "Unable to stat /dev/raw: %s\n",
	    strerror(errno));
    return -errno;
  }
  if(S_ISDIR(stat_buf.st_mode)){  /* The new way */
    new_raw_method = 1;
    if((raw_fd = open("/dev/rawctl", O_RDWR)) < 0){
      fprintf(stderr, "Unable to open raw control device: %s\n",
	      strerror(errno));
      return -errno;
    }
  } else {  /* The old way */
    new_raw_method = 0;
    if((raw_fd = open("/dev/raw", O_RDWR)) < 0){
      fprintf(stderr, "Unable to open raw control device: %s\n",
	      strerror(errno));
      return -errno;
    }
  }

  for(i=64; i<256; i++){
    raw_req.raw_minor = i;
    if(ioctl(raw_fd, RAW_GETBIND, &raw_req)< 0){
      fprintf(stderr, "Unable to perform ioctl on raw control device: %s\n",
	      strerror(errno));
      close(raw_fd);
      return -errno;
    }
    if(raw_req.block_major == major && raw_req.block_minor == minor){
      goto already_bound;
    } else if(!raw_req.block_major && !raw_req.block_minor){
      if(!found)
	found = i;
    }
  }
  if(!found){
    fprintf(stderr, "All raw devices are in use.\n");
    close(raw_fd);
    return -EBUSY;
  }
  raw_req.raw_minor = found;
  raw_req.block_major = major;
  raw_req.block_minor = minor;
  if(ioctl(raw_fd, RAW_SETBIND, &raw_req)< 0){
    fprintf(stderr, "Unable to bind %s to a raw device: %s\n",
	    dev, strerror(errno));
    close(raw_fd);
    return -errno;
  }
 already_bound:
  close(raw_fd);
  if(new_raw_method){
    if(snprintf(rdev, 32, "/dev/raw/raw%d", raw_req.raw_minor) < 0){
      fprintf(stderr, "Programmer error, no enough space for return value.\n");
      return -1;
    }
  } else {
    if(snprintf(rdev, 32, "/dev/raw%d", raw_req.raw_minor) < 0){
      fprintf(stderr, "Programmer error, no enough space for return value.\n");
      return -1;
    }
  }
  
  if(stat(rdev, &stat_buf)){
    if(errno == ENOENT){
      /* Create device node */
      printf("The raw device (162,%d) does not exist.  Creating...\n",
	     raw_req.raw_minor);
      if(mknod(rdev,  S_IFCHR | S_IREAD | S_IWRITE,
	       osi_make_dev(162, raw_req.raw_minor))){
	fprintf(stderr, "Unable to make raw device node: %s\n",
		strerror(errno));
	return -errno;
      }
    } else {
      fprintf(stderr, "Unable to stat %s: %s\n",
	      rdev, strerror(errno));
      return -errno;
    }
  }
  on_exit(_unbind_raw_device, (void *)dev);
  
  return 0;
}


/**
 * get_ip_for_name - 
 * @name: 
 * @ip: 
 * 
 * Another function stolen from Tilsta.  Does what it says.
 *
 * Returns: 0 on success, -1 on error
 */
static int get_ip_for_name(char *name, uint32 *ip)
{
   struct hostent *he;

   he = gethostbyname(name);
   if( he == NULL ) return -1;
   *ip = (uint32)(*(uint32 *)(he->h_addr_list[0]));
   return 0;
}


/*
 * Returns: 1 if valid CCA, 0 otherwise
 */
static int is_valid_local(char *local){
  FILE *local_fp;
  char line[256];

  local_fp = fopen(local, "r");
  if(!local_fp){
    log_sys_err("Unable to open %s", local);
    return 0;
  }
  
  if(!fgets(line, sizeof(line), local_fp)){
    log_sys_err("Failed to read %s", local);
    fclose(local_fp);
    return 0;
  }

  if(strncmp(line, "#CCA:", 5)){
    log_err("%s is not a valid CCS archive.\n", local);
    fclose(local_fp);
    return 0;
  }    
    
  fclose(local_fp);
  return 1;
}


/*
 * is_valid_device:
 * @device:
 *
 * Returns: 1 if a valid device, 0 otherwise
 */
static int is_valid_device(char *device){
  int fd, error=0;
  unsigned long remainder;
  ccs_dh_t header;
  struct stat stat_buf;
  char *sector, *sector_orig;

  sector_orig = (char *)malloc(DEFAULT_BBS + PAGE_SIZE);
  if(!sector_orig){
    fprintf(stderr, "Not enough memory.\n");
    return -1;
  }
  remainder = ((unsigned long)sector_orig) & (PAGE_SIZE-1);
  sector = sector_orig + (PAGE_SIZE - remainder);


  if(stat(device, &stat_buf)){
    if(errno == ENOENT){
      log_err("%s: No such device.\n", device);
    } else {
      log_sys_err("%s", device);
    }
    goto fail;
  } else if(S_ISBLK(stat_buf.st_mode)){
    char raw_device[32];
    if(bind_to_raw_device(device, raw_device) < 0){
      return -1;
    }
    fd = open(raw_device, O_RDONLY/* | O_DIRECT*/);
    if(fd < 0){
      log_sys_err("Unable to open %s (%s)", device, raw_device);
      goto fail;
    }
    if((error = read(fd, sector, DEFAULT_BBS)) != DEFAULT_BBS){
      if(error < 0){
	log_sys_err("Unable to complete read on %s", device);
      } else {
	log_err("Unable to complete read on %s\n", device);
      }
      close(fd);
      goto fail;;
    }
    ccs_dh_in(&header, sector);
    close(fd);

    if(header.dh_magic != CCS_DH_MAGIC){
      log_err("%s is not a valid CCS archive device.\n", device);
      goto fail;
    }

    /* for the future
    if(header.dh_format != CCS_FORMAT_X){
      log_err("incompatible versions..."
      goto fail;
    }
    */
    free(sector_orig);
    return 1;
  } else {
    log_err("%s is not a block device.\n", device);
  }

 fail:
  free(sector_orig);
  return 0;
}

/*
 * Returns: 1 if communication with server is good, 0 otherwise
 */
static int is_valid_server(char *server){
  int fd, pid, wait_stat,i, amount;
  struct sockaddr_in addr;

  char buffer[512];
  struct iovec iov[2];
  ccs_fth_t header;
    
  iov[0].iov_base = &header;
  iov[0].iov_len  = sizeof(header);
  iov[1].iov_base = buffer;
  iov[1].iov_len  = sizeof(buffer);

  header.fth_magic = 0x101674;
  header.fth_last  = 0;
  header.fth_error = 0;

  memset(buffer, 0, sizeof(buffer));

  addr.sin_family = AF_INET;
  addr.sin_port = htons(be_port);
  if(get_ip_for_name(server, (uint32 *)&(addr.sin_addr))){
    log_err("Unable to contact \"%s\": unknown host\n", server);
    return 0;
  }
  /*
  inet_aton(server, &(addr.sin_addr));
  */
  pid = fork();

  if(pid < 0){
    log_sys_err("Unable to create child process");
    return 0;
  } else if(pid > 0){ /* parent */
    for(i=12; i > 0 && !waitpid(pid, &wait_stat, WNOHANG); i--){
      if(i == 10){
	log_msg("Looking for server.");
      } else if(i < 10){
	log_msg(".");
      }
      fflush(NULL);
      sleep(1);
    }
    if(WIFEXITED(wait_stat)){
      if(!WEXITSTATUS(wait_stat)){
	if(i<=10)log_msg("\nServer found.\n");
	return 1;
      } else {
	log_err("\nServer not found.\n");
	return 0;
      }
    } else {
      kill(pid, SIGINT);
      log_err("\nServer not found.\n");
      return 0;
    }
  } else { /* child */
    fd = socket(PF_INET, SOCK_STREAM, 0);

    if(fd < 0){
      log_sys_err("Unable to create socket");
      exit(EXIT_FAILURE);
    }

    if(connect(fd, (struct sockaddr *)&addr, sizeof(addr)) < 0){
      close(fd);
      log_err("ccs_servd is not available on %s\n", server);
      exit(EXIT_FAILURE);
    }
    if(write(fd, "brassow/list_clusters",
	     strlen("brassow/list_clusters")+1) < 0){
      log_sys_err("Failed to write to socket");
      close(fd);
      exit(EXIT_FAILURE);
    }
    amount = readv(fd, iov, 2);
    if((amount < sizeof(header)) || 
       (header.fth_magic != 0x101674) ||
       (header.fth_error) ||
       (header.fth_count + sizeof(header) != amount)){
      if(amount < 0){
	log_sys_err("Unable to communicate with the CCS server ");
      } else if(amount >= sizeof(header)){
	if(read(fd, buffer+(amount-sizeof(header)),
		header.fth_count-(amount-sizeof(header))) ==
	   header.fth_count-(amount-sizeof(header))){
	  goto ok_i_guess;
	}
	log_err("Unable to complete a read from the CCS server.\n");
      }
      close(fd);
      exit(EXIT_FAILURE);
    } else {
      char *tmp;
      int found;
    ok_i_guess:
      found = 0;
      if(header.fth_count == 0){
	log_err("There are no available cluster archives on the server.\n");
	close(fd);
	exit(EXIT_FAILURE);
      }
      tmp= buffer;
      while((tmp = strstr(tmp, cluster_name))){
	if(tmp[strlen(cluster_name)] == '\n'){
	  found = 1;
	  break;
	} else {
	  tmp++;
	}
      }
      if(!found){
	if(!strcmp(cluster_name, "__sentinel__")){
	  printf("%s", buffer);
	} else {
	  log_err("Invalid cluster name, %s.\n", cluster_name);
	  close(fd);
	  exit(EXIT_FAILURE);
	}
      }
    }
    close(fd);
    exit(EXIT_SUCCESS);
  }
}


static int process_local_file(int afd, char request[]){
  int fd = -1;
  int amount;
  char buffer[512];
  ccs_fth_t header;
  struct iovec iov[2];

  /* We can not allow this function to proceed until security issues **
  ** are resolved.                                                   */
  log_err("Refusing to process local file request.\n");
  return -1;

  header.fth_magic = 0x101674;
  header.fth_error = 0;
  header.fth_last = 0;

  iov[0].iov_base = &header;
  iov[0].iov_len  = sizeof(header);
  iov[1].iov_base = buffer;
  /* iov[1].iov_len  = assigned later */

  fd = open(request, O_RDONLY);
  if(fd < 0)
    goto fail;

  while((amount = read(fd, buffer, sizeof(buffer))) >= 0){
    header.fth_count = iov[1].iov_len = amount;
    if(amount < sizeof(buffer))
      header.fth_last = 1;
    if(ccs_writev(afd, iov, 2, 10)){
      log_err("Unable to complete write to socket.");
      goto fail;
    }
    if(header.fth_last){
      close(fd);
      return 0;
    }
  }
 fail:
  if(fd > -1) close(fd);
  header.fth_error = 1;
  header.fth_last = 1;
  header.fth_count = 0;
  write(afd, (char *)&header, sizeof(header));
  return -1;
}


static int process_local_request(int afd){
  FILE *local_fp=NULL;
  int found=0;
  int alternate_request=0;
  char line[256];
  char request[50];
  char buffer[512];

  struct iovec iov[2];
  ccs_fth_t header;
    
  iov[0].iov_base = &header;
  iov[0].iov_len  = sizeof(header);
  iov[1].iov_base = buffer;

  header.fth_magic = 0x101674;
  header.fth_last  = 0;
  header.fth_error = 0;

  memset(request, 0, sizeof(request));

  read(afd, request, sizeof(request));
  request[sizeof(request)-1] = '\0';

  if(verbose){
    log_msg("Request for \"%s\" recieved.\n", request);
  } else {
    log_dbg("Request for \"%s\" recieved.\n", request);
  }
  if(request[0] == '/'){
    log_dbg("Request is for a local file.\n");
    return process_local_file(afd, request);
  }
  if(strstr(request, "/")){
    log_err("Unable to retrieve \"%s\".\n '/'"
	    " is an illegal character.\n", request);
    goto fail;
  }

  local_fp = fopen(local, "rw");
  if(!local_fp){
    log_err("Unable to open local CCS archive, %s.\n", local);
    goto fail;
  }

  memset(buffer, 0, sizeof(buffer));

  while(fgets(line, sizeof(line), local_fp)){
    uint mtime=0;
    uint size=0;
    char filename[256];
    char filename_cmp[256];
    if((sscanf(line, "#nigeb=%s mtime=%u size=%u\n",
	       filename, &mtime, &size) == 3)){
      int diff, amount;
      
      if(!strcmp(request, "list_files")){
	/* fill buffer with file names */
	alternate_request=1;
	if(strlen(filename)+strlen(buffer) > sizeof(buffer)){
	  /* for now, error out.  However, should start new buffer */
	  goto fail;
	}
	sprintf(buffer+strlen(buffer), "%s\n", filename); 
	log_dbg("Filling request buffer: %d, %s\n", strlen(buffer), buffer);
	continue;
      } else if(strcmp(request, filename)){
	/* Not the file we are looking for */
	continue;
      }
      
      found=1;
      diff=size;
      amount=(diff < sizeof(buffer))?diff:sizeof(buffer);

      while(1){
	memset(buffer, 0, sizeof(buffer));
	if(fread(buffer, sizeof(char), amount, local_fp) != amount){
	  log_err("Unable to read file, %s from local CCS archive.\n",
		  request);
	  goto fail;
	}
	diff -= amount;

	iov[1].iov_len = header.fth_count = amount;

	if(diff){
	  if(ccs_writev(afd, iov, 2, 10)){
	    log_err("Unable to complete write to socket.\n");
	    goto fail;
	  }
	} else {
	  if(!fgets(line, sizeof(line), local_fp) ||
	     !sscanf(line, "#dne=%s\n", filename_cmp) ||
	     strcmp(filename, filename_cmp)){
	    goto fail;
	  } else {
	    header.fth_last=1;
	    if(ccs_writev(afd, iov, 2, 10)){
	      log_err("Unable to complete write to socket.\n");
	      goto fail;
	    }
	    goto out;
	  }
	}
	amount=(diff < sizeof(buffer))?diff:sizeof(buffer);
      }
    }
  }

  if(alternate_request){
    found = 1;
    header.fth_magic = 0x101674;
    header.fth_count = strlen(buffer)+1;
    header.fth_last  = 1;
    header.fth_error = 0;
    iov[1].iov_len = header.fth_count;

    if(ccs_writev(afd, iov, 2, 10)){
      log_err("Unable to complete write to socket.\n");
      goto fail;
    }
  }  

  if(!found){
    header.fth_magic = 0x101674;
    header.fth_count = 0;
    header.fth_last  = 1;
    header.fth_error = 0;

    if(verbose){
      log_msg("Requested file not found :: %s\n", request);
    }
    write(afd, (char *)&header, sizeof(header));
  }

 out:
  fclose(local_fp);
  if(verbose){
    log_msg("Completed request for \"%s\".\n", request);
  } else {
    log_dbg("Completed request for \"%s\".\n", request);
  }
  return 0;

 fail:
  log_err("Failed on request for \"%s\".\n", request);
  header.fth_magic = 0x101674;
  header.fth_count = 0;
  header.fth_last  = 1;
  header.fth_error = 1;

  write(afd, (char *)&header, sizeof(header));

  if(local_fp) fclose(local_fp);

  return -1;
}


static int process_server_request(int afd){
  int be_sfd = -1;
  char request[50];
  char *request_filename = NULL;
  struct sockaddr_in addr;
  ccs_fth_t header;

  memset(request, 0, sizeof(request));

  read(afd, request, sizeof(request));
  /* truncate any requests that are too large */
  request[sizeof(request)-1] = '\0';

  if(verbose){
    log_msg("Request for \"%s\" received.\n", request);
  } else {
    log_dbg("Request for \"%s\" received.\n", request);
  }

  if(request[0] == '/'){
    log_dbg("Request is for a local file.\n");
    return process_local_file(afd, request);
  }

  if(strstr(request, "/")){
    /* the '/' character is not allowed in the filename */
    log_err("Unable to retrieve \"%s\".\n '/'"
	    " is an illegal character.\n", request);
    goto fail;
  }

  /* +2 because we need space for trailing NULL and '/' char */
  request_filename = malloc(strlen(request)+strlen(cluster_name)+2);
  if(!request_filename){
    log_err("Can't malloc request_filename.\n");
    goto fail;
  }

  sprintf(request_filename, "%s/%s", cluster_name, request);

  be_sfd = socket(PF_INET, SOCK_STREAM, 0);

  if(be_sfd < 0){
    log_sys_err("Unable to create socket");
    goto fail;
  }

  addr.sin_family = AF_INET;
  addr.sin_port = htons(be_port);
  if(server){
    if(get_ip_for_name(server, (uint32 *)&(addr.sin_addr))){
      goto fail;
    }
  } else {
    addr.sin_addr.s_addr = INADDR_ANY;
  }

  if(connect(be_sfd, (struct sockaddr *)&addr, sizeof(addr)) < 0){
    log_sys_err("Unable to connect to server");
    close(be_sfd);
    goto fail;
  } else {
    /* Request file from back end */
    if(write(be_sfd, request_filename, strlen(request_filename)) < 0){
      log_sys_err("Failed to write to socket");
      goto fail;
    }

    if(verbose){
      log_msg("Request for \"%s\" sent to server.\n",
	      request_filename);
    } else {
      log_dbg("Request for \"%s\" sent to server.\n",
	      request_filename);
    }
  }

  while(1){
    int amount;
    char buffer[512];
    struct iovec iov[2];

    iov[0].iov_base = &header;
    iov[0].iov_len  = sizeof(header);
    iov[1].iov_base = buffer;
    iov[1].iov_len  = sizeof(buffer);

    memset(&header, 0, sizeof(header));
    memset(buffer, 0, sizeof(buffer));
    /* Recieve file from back end */
    amount = readv(be_sfd, iov, 2);
    if(amount < 0){
      log_sys_err("Read from socket failed");
      goto fail;
    } else if(amount < sizeof(header)){
      log_err("Incomplete header received from server.\n");
      goto fail;
    } else if(header.fth_magic != 0x101674 || header.fth_error){
      log_err("Bad header received:\n"
	      "fth_magic = 0x%x, fth_error = %d\n",
	      header.fth_magic, header.fth_error);
      goto fail;
    }
    if((amount-sizeof(header)) != header.fth_count){
      /* Sometimes, we read everything that was on the wire, but it **
      ** isn't everything that was sent.  So, we try to finish the  **
      ** job here.  We put a limit on the number of retries and we  **
      ** don't tolerate read failures...                            */
      int rtn;
      int try_again=5;
      int index = amount-sizeof(header);
      int diff = header.fth_count - (amount-sizeof(header));

      for(; try_again && diff; try_again--){
	rtn = read(be_sfd, &(buffer[index]), diff);
	if(rtn < 0){
	  log_err("Failed to complete read:\n"
		      "diff = %d, rtn = %d\n",
		      diff, rtn);
	  goto fail;
	}
	diff -= rtn;
	index += rtn;
      }

      if(diff){
	log_err("Failed to complete read:\n"
		"diff = %d\n",
		diff);
	goto fail;
      }
    }
    iov[1].iov_len  = header.fth_count;
    if(ccs_writev(afd, iov, 2, 10)){
      log_err("Unable to complete write to socket.\n");
      goto fail;
    }

    if(header.fth_last){
      break;
    }
  }

  free(request_filename);

  close(be_sfd);
  if(verbose){
    log_msg("Completed request for \"%s\".\n", request);
  } else {
    log_dbg("Completed request for \"%s\".\n", request);
  }
  return 0;

 fail:
  log_err("Failed on request for \"%s\".\n", request_filename);
  if(request_filename)
    free(request_filename);

  header.fth_magic = 0x101674;
  header.fth_count = 0;
  header.fth_last  = 1;
  header.fth_error = 1;

  write(afd, (char *)&header, sizeof(header));
  if(be_sfd > 0) close(be_sfd);
  return -1;
}


static int process_device_request(int afd){
  int file_method = 0;
  int tmp_fd=-1;
  int found=0;
  int alternate_request=0;
  unsigned long remainder;
  FILE *tmp_fp=NULL;
  char tmp_file[100];
  char line[256];
  char request[50];
  char *sector, *sector_orig;
  char buffer[512];
  char *archive_data=NULL;
  char *char_ptr=NULL;
  ccs_dh_t dev_header;

  struct iovec iov[2];
  ccs_fth_t header;
    
  iov[0].iov_base = &header;
  iov[0].iov_len  = sizeof(header);
  iov[1].iov_base = buffer;

  header.fth_magic = 0x101674;
  header.fth_last  = 0;
  header.fth_error = 0;

  memset(request, 0, sizeof(request));

  read(afd, request, sizeof(request));
  request[sizeof(request)-1] = '\0';

  if(verbose){
    log_msg("Request for \"%s\" recieved.\n", request);
  } else {
    log_dbg("Request for \"%s\" recieved.\n", request);
  }
  if(request[0] == '/'){
    log_dbg("Request is for a local file.\n");
    return process_local_file(afd, request);
  }

  sector_orig = (char *)malloc(DEFAULT_BBS + PAGE_SIZE);
  if(!sector_orig){
    fprintf(stderr, "Not enough memory.\n");
    return -1;
  }
  remainder = ((unsigned long)sector_orig) & (PAGE_SIZE-1);
  sector = sector_orig + (PAGE_SIZE - remainder);

  if(strstr(request, "/")){
    log_err("Unable to retrieve \"%s\".\n '/'"
	    " is an illegal character.\n", request);
    goto fail;
  }

  if(lseek(device_fd, 0, SEEK_SET) < 0){
    log_sys_err("Bad seek on device, \"%s\"", device);
    goto fail;
  }
  
  if(read(device_fd, sector, DEFAULT_BBS) < DEFAULT_BBS){
    log_err("Bad read on device, \"%s\"\n", device);
    goto fail;
  }

  ccs_dh_in(&dev_header, sector);
  if(dev_header.dh_magic == CCS_DH_MAGIC){
    if(!dev_header.dh_active){
      if(lseek(device_fd, dev_header.dh_maxsize, SEEK_SET) < 0){
	log_sys_err("Bad seek on device, \"%s\"", device);
	goto fail;
      }
    }
  } else {
    log_err("The device, \"%s\", does not contain a CCS archive.\n",
	    device);
    goto fail;
  }

  if((dev_header.dh_size > 4*1024*1024) || !(archive_data = malloc(dev_header.dh_size))){
    log_dbg("Going with tmp file extraction.\n");
    file_method = 1;
    sprintf(tmp_file, "/var/run/sistina/ccsd_tmp_XXXXXX");
    
    tmp_fd = mkstemp(tmp_file);

    if(tmp_fd < 0){
      log_err("Unable to create temporary achive.\n");
      goto fail;
    }

    unlink(tmp_file);
  } else {
    log_dbg("%d bytes of memory allocated for archive.\n", dev_header.dh_size);
  }

  if(file_method){
    while(dev_header.dh_size){
      int write_size;
      if(read(device_fd, sector, DEFAULT_BBS) < DEFAULT_BBS){
	log_err("Bad read on device, \"%s\"\n", device);
	goto fail;
      }
      write_size = (dev_header.dh_size < DEFAULT_BBS) ? 
	dev_header.dh_size : DEFAULT_BBS;
      if(write(tmp_fd, sector, write_size) < write_size){
	log_sys_err("Unable to write to temporary archive");
	goto fail;
      }
      dev_header.dh_size -= write_size;
    }

    if(lseek(tmp_fd, 0, SEEK_SET)){
      log_err("Unable to reset offset in temporary archive.\n");
      goto fail;
    }
  
    tmp_fp = fdopen(tmp_fd, "rw");
    if(!tmp_fp){
      log_err("Unable to reopen temporary archive.\n");
      goto fail;
    }

    memset(buffer, 0, sizeof(buffer));

    while(fgets(line, sizeof(line), tmp_fp)){
      uint mtime=0;
      uint size=0;
      char filename[256];
      char filename_cmp[256];
      if((sscanf(line, "#nigeb=%s mtime=%u size=%u\n",
		 filename, &mtime, &size) == 3)){
	int diff, amount;

	if(!strcmp(request, "list_files")){
	  /* fill buffer with file names */
	  alternate_request=1;
	  if(strlen(filename)+strlen(buffer) > sizeof(buffer)){
	    /* for now, error out.  However, should start new buffer */
	    goto fail;
	  }
	  sprintf(buffer+strlen(buffer), "%s\n", filename); 
	  log_dbg("Filling request buffer: %d, %s\n", strlen(buffer), buffer);
	  continue;
	} else if(strcmp(request, filename)){
	  /* Not the file we are looking for */
	  continue;
	}
      
	found = 1;
	diff=size;
	amount=(diff < sizeof(buffer))?diff:sizeof(buffer);
	
	while(1){
	  memset(buffer, 0, sizeof(buffer));
	  if(fread(buffer, sizeof(char), amount, tmp_fp) != amount){
	    log_err("Unable to read file, %s from temporary archive.\n",
		    request);
	    goto fail;
	  }
	  diff -= amount;

	  iov[1].iov_len = header.fth_count = amount;

	  if(diff){
	    if(ccs_writev(afd, iov, 2, 10)){
	      log_err("Unable to complete write to socket.\n");
	      goto fail;
	    }
	  } else {
	    if(!fgets(line, sizeof(line), tmp_fp) ||
	       !sscanf(line, "#dne=%s\n", filename_cmp) ||
	       strcmp(filename, filename_cmp)){
	      goto fail;
	    } else {
	      header.fth_last=1;
	      if(ccs_writev(afd, iov, 2, 10)){
		log_err("Unable to complete write to socket.\n");
		goto fail;
	      }
	      goto out;
	    }
	  }
	  amount=(diff < sizeof(buffer))?diff:sizeof(buffer);
	}
      }
    }
  } else {  /* !file_method */
    char_ptr = archive_data;
    while(dev_header.dh_size){
      int write_size;
      if(read(device_fd, sector, DEFAULT_BBS) < DEFAULT_BBS){
	log_err("Bad read on device, \"%s\"\n", device);
	goto fail;
      }
      write_size = (dev_header.dh_size < DEFAULT_BBS) ? 
	dev_header.dh_size : DEFAULT_BBS;

      memcpy(char_ptr, sector, write_size);
      char_ptr += write_size;
      dev_header.dh_size -= write_size;
    }

    memset(buffer, 0, sizeof(buffer));
  
    char_ptr = archive_data;
    while((char_ptr = strstr(char_ptr, "#nigeb="))){
      uint mtime=0;
      uint size=0;
      char filename[256];
      char filename_cmp[256];

      if((sscanf(char_ptr, "#nigeb=%s mtime=%u size=%u\n",
		 filename, &mtime, &size) == 3)){
	int diff, amount;

	char_ptr = strstr(char_ptr, "\n")+1;
	if(!strcmp(request, "list_files")){
	  /* fill buffer with file names */
	  alternate_request=1;
	  if(strlen(filename)+strlen(buffer) > sizeof(buffer)){
	    /* for now, error out.  However, should start new buffer */
	    goto fail;
	  }
	  sprintf(buffer+strlen(buffer), "%s\n", filename); 
	  log_dbg("Filling request buffer: %d, %s\n", strlen(buffer), buffer);
	  continue;
	} else if(strcmp(request, filename)){
	  /* Not the file we are looking for */
	  continue;
	}

	found = 1;
	diff=size;
	amount=(diff < sizeof(buffer))?diff:sizeof(buffer);

	while(1){
	  memset(buffer, 0, sizeof(buffer));
	  memcpy(buffer, char_ptr, amount);
	  char_ptr += amount;
	  diff -= amount;

	  iov[1].iov_len = header.fth_count = amount;

	  if(diff){
	    if(ccs_writev(afd, iov, 2, 10)){
	      log_err("Unable to complete write to socket.\n");
	      goto fail;
	    }
	  } else {
	    if(!sscanf(char_ptr, "#dne=%s\n", filename_cmp) ||
	       strcmp(filename, filename_cmp)){
	      goto fail;
	    } else {
	      header.fth_last=1;
	      if(ccs_writev(afd, iov, 2, 10)){
		log_err("Unable to complete write to socket.\n");
		goto fail;
	      }
	      goto out;
	    }
	  }
	  amount=(diff < sizeof(buffer))?diff:sizeof(buffer);
	}
      } else {
	char_ptr = strstr(char_ptr, "\n")+1;
      }
    }
  }

  if(alternate_request){
    found = 1;
    header.fth_magic = 0x101674;
    header.fth_count = strlen(buffer)+1;
    header.fth_last  = 1;
    header.fth_error = 0;
    iov[1].iov_len = header.fth_count;

    if(ccs_writev(afd, iov, 2, 10)){
      log_err("Unable to complete write to socket.\n");
      goto fail;
    }
  }  

  if(!found){
    header.fth_magic = 0x101674;
    header.fth_count = 0;
    header.fth_last  = 1;
    header.fth_error = 0;

    log_msg("Requested file not found :: %s\n", request);

    write(afd, (char *)&header, sizeof(header));
  }

 out:
  if(archive_data) free(archive_data);
  if(tmp_fp) fclose(tmp_fp);
  free(sector_orig);

  if(verbose){
    log_msg("Completed request for \"%s\".\n", request);
  } else {
    log_dbg("Completed request for \"%s\".\n", request);
  }

  return 0;

 fail:
  header.fth_magic = 0x101674;
  header.fth_count = 0;
  header.fth_last  = 1;
  header.fth_error = 1;

  log_err("Failed on request for \"%s\".\n", request);
  write(afd, (char *)&header, sizeof(header));

  /* tmp_fp was opened from tmp_fd.  So, closing tmp_fp will close tmp_fd */
  if(tmp_fp) fclose(tmp_fp);
  else if(tmp_fd>=0) close(tmp_fd);

  if(archive_data) free(archive_data);
  free(sector_orig);
  return -1;      
}
