/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**  
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>

#include <cluster/fence_driver_msg.h>
#include "fence_log.h"

#define FENCE_DEVICE_NAME     "/dev/fence"

/*
 * check_fence_module() - check if fence module is loaded.
 *
 * This function will check to make sure the fence module
 * is loaded.  If the module is loaded, the minor number
 * is returned.
 *
 * Returns: minor on success, -1 on error
 */
static int check_fence_module(void){
  FILE *pf;
  int minor, error;
  unsigned char name[32];
  
  /* check if fence module is loaded in kernel */
  
  pf = fopen("/proc/misc", "r");
  if (!pf)
  {
    log_err("Can't open /proc/misc\n");
    return -1;
  }
  
  while ((error = fscanf(pf, "%u %s\n", &minor, name)) != EOF)
    if (!strncmp(name, "fence", 5))
      break;
  
  if (error == EOF)
  {
    log_err("No \"fence\" found in /proc/misc\n");
    return -1;
  }
  
  fclose(pf);
  return minor;
}  


/*
 * initialize_fence_device - initialize/open the fence device
 *
 * This function will check to make sure the fence module is
 * loaded.  It will also create the device node to interface
 * with the module.  It then opens the device node and returns
 * the file descriptor.
 *
 * Returns: fd on success, -1 on error
 */
int initialize_fence_device(void){
  int error, minor, fd;
  
  minor = check_fence_module();

  if(minor < 0){
    return -1;
  }

  /* create proper device node */
    
  unlink(FENCE_DEVICE_NAME);
  error = mknod(FENCE_DEVICE_NAME, 0600 | S_IFCHR, makedev(10, minor));
  if (error < 0){
    log_err("Can't make device node, %s\n", FENCE_DEVICE_NAME);
    return -1;
  }
    
  /* fence requests come from /dev/fence */
  /* read and write fds for the fence device are the same */
  fd = open(FENCE_DEVICE_NAME, O_RDWR);
  if (fd < 0){
    log_err("Can't open %s\n", FENCE_DEVICE_NAME);
    return -1;
  }
  
  return fd;
}


/*
 * get_fence_device_request - read fence char device for fence requests
 *
 * This function will read the character device implemented by
 * the fence kernel module and returns the name of the node to
 * be fenced.  THE RETURNED VALUE MUST BE FREED BY THE CALLER
 * OF THIS FUNCTION.
 *
 * Returns: name of victim on success, NULL on error
 */
char *get_fence_device_request(int fd){
  int error=0;
  char buffer[FENCE_READBUF_SIZE];
  char *victim;


  error = (lseek(fd, 0, SEEK_SET) != 0);
  if(error){
    log_err("Unable to seek on fence device, %s\n", FENCE_DEVICE_NAME);
    return NULL;
  }

  error = read(fd, buffer, FENCE_READBUF_SIZE);
  if(error != FENCE_READBUF_SIZE){
    log_err("Unable to read fence device, %s\n", FENCE_DEVICE_NAME);
    return NULL;
  }

  victim = (char *)malloc(strlen(buffer)+1);
  if(!victim){
    return NULL;
  }
  strcpy(victim, buffer);
  return victim;
}


/*
 * report_fence_device_result - return fence result to fence dev
 * @fd: file desciptor of the fence character device
 * @result:  1 = success, 0 = failure
 *
 * This function will try to write the result of the fence operation
 * to the fence character device.
 *
 * Returns: 0 on success, -1 on failure
 */
int report_fence_device_result(int fd, int result){
  int error;
  char msg[FENCE_WRITEBUF_SIZE];

  error = lseek(fd, 0, SEEK_SET);
  if (error){
    log_err("can't seek on fence device, %s.\n",FENCE_DEVICE_NAME);
    return -1;
  }

  if (result)
    strcpy(msg, FENCE_MSG_OK);
  else
    strcpy(msg, FENCE_MSG_FAIL);

  error = write(fd, msg, sizeof(msg));

  if (error != sizeof(msg)){
    log_err("Unable to write to fence device, %s\n", FENCE_DEVICE_NAME);
    return -1;
  }

  return 0;
}

