/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**  
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <syslog.h>
#include <signal.h>

#include "copyright.cf"
#include "fence_release.cf"
#include "utils.h"
#include "fence_lockfile.h"
#include "fence_log.h"
#include "fence_dev.h"
#include <cluster/ccs.h>

#define FENCED_LOCKFILE "/var/run/sistina/fenced.pid"

static void print_usage(FILE *stream);
static void _destroy_lockfile(void);
static int daemonize(void);

int main(int argc, char *argv[]){
  int c;
  int daemon = 1;
  int fence_fd = -1;
  int error=0;
  char *victim;

  while((c = getopt(argc, argv, "hnV")) != -1){
    switch(c){
    case 'h':
      print_usage(stdout);
      exit(EXIT_SUCCESS);

    case 'n':
      daemon = 0;
      break;

    case 'V':
      printf("%s %s (built %s %s)\n", argv[0], FENCE_RELEASE_NAME,
	     __DATE__, __TIME__);
      printf("%s\n", SISTINA_COPYRIGHT);
      exit(EXIT_SUCCESS);
    default:
      print_usage(stderr);
      exit(EXIT_FAILURE);
    }
  }
  
  fence_fd = initialize_fence_device();
  if(fence_fd < 0){
    log_err("Unable to initialize the fence device.\n"
	    "Hint:  Make sure the \"fence\" module is loaded.\n");
    exit(EXIT_FAILURE);
  }

  if(daemon){
    if(daemonize()){
      exit(EXIT_FAILURE);
    }
  }

  while(1){
    if((victim = get_fence_device_request(fence_fd)) == NULL){
      log_err("Error while reading fence device.\n");
    } else {
      log_dbg("Fence request for \"%s\" recieved from the fence device.\n",
	      victim);

      error = process_fence_request(victim);

      if(error){
	if(report_fence_device_result(fence_fd, 0)){
	  log_err("Unable to report failed result of the fence.\n");
	} else {
	  log_err("Fence of \"%s\" was unsuccessful.\n", victim);
	}
      } else {
	if(report_fence_device_result(fence_fd, 1)){
	  log_err("Unable to report successful result of the fence.\n");
	} else {
	  log_msg("Fence of \"%s\" was successful.\n", victim);
	}
      }
      free(victim);
    }
  }

  exit(EXIT_SUCCESS);
}


static void print_usage(FILE *stream){
  fprintf(stream,
	  "Usage:\n"
	  "\n"
	  "fenced [options]\n"
	  "\n"
	  "Options:\n"
	  "  -h        Help.\n"
	  "  -n        No Daemon.  Run in the foreground.\n"
	  "  -V        Version information.\n"
	  );
}


static void _destroy_lockfile(void){
  destroy_lockfile(FENCED_LOCKFILE);
}


static int daemonize(void){
  int pid;

  if(test_lockfile(FENCED_LOCKFILE)){
    return -1;
  }

  pid = fork();

  if(pid<0){
    log_err("Unable to fork.\n");
    return -1;
  } else if(pid){
    exit(EXIT_SUCCESS);
  }

  setsid();
  chdir("/");
  umask(0);

  if(create_lockfile(FENCED_LOCKFILE)){
    log_err("Unable to create lock file.\n");
    return -1;
  }

  close(0); close(1); close(2);
  open("/dev/null", O_RDONLY);  /* reopen stdin */
  open("/dev/null", O_WRONLY);  /* reopen stdout */
  open("/dev/null", O_WRONLY);  /* reopen stderr */    

  atexit(&_destroy_lockfile);
  signal(SIGINT, &exit); signal(SIGQUIT, &exit); signal(SIGTERM, &exit);
  openlog("fenced", LOG_PID, LOG_DAEMON);
  fence_syslog_enable();
  return 0;
}
