/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include "servd.h"
#include "trans.h"
#include <linux/gnbd_request.h>
#include "device.h"
#include "gserv.h"
#include "fence.h"

LIST_HEAD(request_list);
struct semaphore request_list_lk;
wait_queue_head_t request_sleep;
atomic_t request_count;

void
init_gnbd_requestd(void)
{
	init_MUTEX(&request_list_lk);
	init_waitqueue_head(&request_sleep);
	atomic_set(&request_count, 0);
}

int
add_to_request_list(u32 client_ip, request_t ** request_info)
{
	int err = 0;
	request_t *info;

	*request_info = NULL;
	info = (request_t *) kmalloc(sizeof (request_t), GFP_KERNEL);
	if (!info) {
		printe("couldn't allocate memory for request_info\n");
		err = -ENOMEM;
		goto out;
	}
	info->task = current;
	info->client_ip = client_ip;
	info->pid = (long) (current->pid);
	down(&request_list_lk);
	list_add(&info->list, &request_list);
	up(&request_list_lk);
	*request_info = info;

      out:
	return err;
}

void
remove_from_request_list(request_t * request_info)
{

	down(&request_list_lk);
	list_del(&request_info->list);
	kfree(request_info);
	up(&request_list_lk);

	return;
}

int
kill_requests(u32 client_ip)
{
	int err = 0;
	list_t *list_item;
	request_t *request_info = NULL;

	down(&request_list_lk);
	list_for_each(list_item, &request_list) {
		request_info = list_entry(list_item, request_t, list);
		if (!client_ip || client_ip == request_info->client_ip) {
			err = send_sig(SIGINT, request_info->task, 0);
			if (err) {
				printe
				    ("error sending SIGINT to gnbd_requestd, pid %ld. (%d)\n",
				     request_info->pid, err);
				up(&request_list_lk);
				goto out;
			}
		}
	}
	up(&request_list_lk);

	while (REQUESTS_LEFT) {
		DECLARE_WAITQUEUE(__wait_chan, current);
		current->state = TASK_INTERRUPTIBLE;
		add_wait_queue(&request_sleep, &__wait_chan);
		if (REQUESTS_LEFT)
			schedule();
		remove_wait_queue(&request_sleep, &__wait_chan);
		current->state = TASK_RUNNING;
	}

      out:
	return err;
}

#define DO_TRANS(action, label) \
do { \
  char __ip_str[16]; \
  err = (action); \
  if (err){ \
    printe("transfer with %s failed at line %d (%d)\n", \
           beip_to_str(client_ip, __ip_str), __LINE__, err); \
    goto label; \
  } \
} while (0)

int
gnbd_requestd(void *args)
{
	int err;
	int free_socket = TRUE;
	request_t *request_info;
	struct socket *sock = (struct socket *) args;
	signal_info_t sig_info;
	u32 client_ip;
	u32 req;
	u32 reply = 0;

	gnbd_daemonize("gnbd_requestd");
	reparent_to_init();
	setup_blocking(&sig_info);
	err = get_peer_ip(sock, &client_ip);
	if (err)
		goto out;

	err = add_to_request_list(client_ip, &request_info);
	if (err)
		goto out;

	if (gnbd_servd_state == GNBD_SERVD_DEAD)
		goto out_list;

	DO_TRANS(timed_transfer_be32(RECV, sock, &req), out_list);
	switch (req) {
	case GNBD_SERV_LOGIN_REQ:
		{
			char ip_str[16];

			INC_GSERV_COUNT;
			err = kernel_thread(gserv, sock, 0);
			if (err < 0) {
				printe("couldn't start gserv thread for %s\n",
				       beip_to_str(client_ip, ip_str));
				break;
			}
			free_socket = FALSE;
			break;
		}
	case GNBD_SERV_CREATE_REQ:
		{
			char *dev = NULL;
			create_req_t create_req;
			DO_TRANS(timed_transfer_info(RECV, sock, &create_req,
						     sizeof (create_req)),
				 fail_create);
			CREATE_REQ_TO_CPU(&create_req);
			dev = (char *) kmalloc(create_req.dev_len, GFP_KERNEL);
			if (!dev) {
				printe
				    ("couldn't allocate memory for server device\n");
				err = -ENOMEM;
				reply = ENOMEM;
				DO_TRANS(transfer_be32(RECV, sock, &reply),
					 fail_create);
				break;
			}
			DO_TRANS(timed_transfer_info(RECV, sock, dev,
						     create_req.dev_len),
				 fail_create);
			dev[create_req.dev_len - 1] = 0;
			err =
			    create_device(create_req.name, dev,
					  create_req.timeout,
					  create_req.readonly);
			if (err < 0)
				reply = -err;
			DO_TRANS(transfer_be32(SEND, sock, &reply),
				 fail_create);
		      fail_create:
			if (dev)
				kfree(dev);
			break;
		}
	case GNBD_SERV_REMOVE_REQ:
	case GNBD_SERV_FORCE_REMOVE_REQ:
		{
			device_req_t remove_req;
			DO_TRANS(timed_transfer_info(RECV, sock, &remove_req,
						     sizeof (remove_req)),
				 out_list);
			if (req == GNBD_SERV_REMOVE_REQ)
				err = remove_device(remove_req.name, 0);
			else
				err = remove_device(remove_req.name, 1);
			if (err < 0)
				reply = -err;
			DO_TRANS(transfer_be32(SEND, sock, &reply), out_list);
			break;
		}
	case GNBD_SERV_TEST_ACTIVE_REQ:
		validate_gserv();
		DO_TRANS(transfer_be32(SEND, sock, &reply), out_list);
		break;
	case GNBD_SERV_FULL_LIST_REQ:
		{
			char *buffer = NULL;
			u32 size;

			err = full_list(&buffer, &size);
			if (err < 0) {
				reply = -err;
				DO_TRANS(transfer_be32(SEND, sock, &reply),
					 out_list);
				break;
			}
			DO_TRANS(transfer_be32(SEND, sock, &reply),
				 fail_full_list);
			DO_TRANS(transfer_be32(SEND, sock, &size),
				 fail_full_list);
			if (size)
				DO_TRANS(transfer_info
					 (SEND, sock, buffer, size),
					 fail_full_list);
		      fail_full_list:
			kfree(buffer);
			break;
		}
	case GNBD_SERV_NAMES_LIST_REQ:
		{
			char *buffer = NULL;
			u32 size;

			err = names_list(&buffer, &size);
			if (err < 0) {
				reply = -err;
				DO_TRANS(transfer_be32(SEND, sock, &reply),
					 out_list);
				break;
			}
			DO_TRANS(transfer_be32(SEND, sock, &reply),
				 fail_names_list);
			DO_TRANS(transfer_be32(SEND, sock, &size),
				 fail_names_list);
			if (size)
				DO_TRANS(transfer_info
					 (SEND, sock, buffer, size),
					 fail_names_list);
		      fail_names_list:
			kfree(buffer);
			break;
		}
	case GNBD_SERV_FENCE_REQ:
		{
			u32 fence_ip;
			DO_TRANS(timed_transfer_info
				 (RECV, sock, &fence_ip, sizeof (u32)),
				 out_list);
			err = add_to_banned_list(fence_ip);
			if (err < 0) {
				reply = -err;
				DO_TRANS(transfer_be32(SEND, sock, &reply),
					 out_list);
				break;
			}
			err = kill_gserv(fence_ip, NULL);
			if (err < 0)
				reply = -err;
			DO_TRANS(transfer_be32(SEND, sock, &reply), out_list);
			break;
		}
	case GNBD_SERV_UNFENCE_REQ:
		{
			u32 unfence_ip;
			u64 timestamp;
			DO_TRANS(timed_transfer_info
				 (RECV, sock, &unfence_ip, sizeof (u32)),
				 out_list);
			DO_TRANS(timed_transfer_info
				 (RECV, sock, &timestamp, sizeof (u64)),
				 out_list);
			timestamp = be64_to_cpu(timestamp);
			err = remove_from_banned_list(unfence_ip, timestamp);
			if (err < 0)
				reply = -err;
			DO_TRANS(transfer_be32(SEND, sock, &reply), out_list);
			break;
		}
	case GNBD_SERV_LIST_BANNED_REQ:
		{
			char *buffer = NULL;
			u32 size;

			err = list_banned(&buffer, &size);
			if (err < 0) {
				reply = -err;
				DO_TRANS(transfer_be32(SEND, sock, &reply),
					 out_list);
				break;
			}
			DO_TRANS(transfer_be32(SEND, sock, &reply),
				 fail_list_banned);
			DO_TRANS(transfer_be32(SEND, sock, &size),
				 fail_list_banned);
			if (size)
				DO_TRANS(transfer_info
					 (SEND, sock, buffer, size),
					 fail_list_banned);
		      fail_list_banned:
			kfree(buffer);
			break;
		}
	case GNBD_SERV_KILL_GSERV:
		{
			device_req_t kill_req;
			DO_TRANS(timed_transfer_info(RECV, sock, &kill_req,
						     sizeof (kill_req)),
				 out_list);
			kill_bad_gserv(client_ip, kill_req.name);
			DO_TRANS(transfer_be32(SEND, sock, &reply), out_list);
			break;
		}
	default:
		reply = -ENOTTY;
		DO_TRANS(transfer_be32(SEND, sock, &reply), out_list);
	}
      out_list:
	remove_from_request_list(request_info);
      out:
	if (free_socket)
		sock_release(sock);
	DEC_REQUEST_COUNT;
	wake_up(&request_sleep);

	return err;
}
