/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/


#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "global.h"
#include "osi_endian.h"
#include "gnbd_release.cf"
#include "copyright.cf"
#include <linux/gnbd_request.h>

#define TIMEOUT_DEFAULT 60

int message_flag = 1;
u16 port = 14243;

#define MAN_MSG   "Please see man page for details.\n"

#define printm(fmt, args...)\
do { \
  if(message_flag != 0) \
    printf("gnbd_export: " fmt, ##args); \
} while(0)

#define printv(fmt, args...)\
do { \
  if(message_flag == 2) \
    printf("gnbd_export: " fmt, ##args); \
} while(0)

#define printe(fmt, args...) fprintf(stderr, "gnbd_export error: " fmt, ##args)

int connect_to_gnbd_servd(){
  struct sockaddr_in server;
  int sock_fd;
  
  sock_fd = socket(AF_INET, SOCK_STREAM, 0);
  if (sock_fd < 0){
    printe("error creating socket: %s\n", strerror(errno));
    return -1;
  }
  if (!inet_aton("127.0.0.1", (struct in_addr *)&server.sin_addr.s_addr)){
    printe("cannot set socket address to 127.0.0.1\n");
    return -1;
  }
  server.sin_port = htons(port);
  server.sin_family = AF_INET;
  if (connect(sock_fd, (struct sockaddr *)&server, sizeof(server)) < 0){
    close(sock_fd);
    printe("error connecting to gnbd_servd: %s\n", strerror(errno));
    return -1;
  }
  return sock_fd;
}

int send_cmd(int fd, u32 cmd, char *type){
  cmd = cpu_to_be32(cmd);
  if (write(fd, &cmd, sizeof(cmd)) != sizeof(u32)){
    printe("sending %s command failed : %s\n", type, strerror(errno));
    return -1;
  }
  return 0;
}

int recv_reply(int fd, char *type){
  u32 reply;
  
  if (read(fd, &reply, sizeof(reply)) != sizeof(reply)){
    printe("reading %s reply failed : %s\n", type, strerror(errno));
    return -1;
  }
  reply = be32_to_cpu(reply);
  if (reply)
    printe("%s request failed : %s\n", type, strerror(reply));
  return -reply;
}

int servcreate(char *name, char *device, u32 timeout, u8 readonly){
  create_req_t create_req;
  int fd;

  strncpy(create_req.name, name, 255);
  create_req.name[255] = 0;
  if (strchr(create_req.name, '/')){
    printe("server name %s is invalid. Names cannot contain a '/'\n",
           create_req.name);
    return 1;
  }
  create_req.dev_len = strlen(device) + 1;
  create_req.timeout = timeout;
  CREATE_REQ_TO_BE(&create_req);
  create_req.readonly = readonly;
  fd = connect_to_gnbd_servd();
  if (fd < 0)
    return 1;
  if (send_cmd(fd, GNBD_SERV_CREATE_REQ, "create") < 0)
    return 1;
  if (write(fd, &create_req, sizeof(create_req)) != sizeof(create_req)){
    printe("sending create data failed : %s\n", strerror(errno));
    return 1;
  }
  if (write(fd, device, strlen(device) + 1) != (strlen(device) + 1)){
    printe("sending device name failed : %s\n", strerror(errno));
    return 1;
  }
  if (recv_reply(fd, "create") < 0)
    return 1;
  printm("created GNBD %s serving file %s\n", name, device);
  close(fd);
  return 0;
}

int servremove(char *name, int force){
  device_req_t remove_req;
  int fd;

  strncpy(remove_req.name, name, 255);
  remove_req.name[255] = 0;
  fd = connect_to_gnbd_servd();
  if (fd < 0)
    return 1;
  if (force){
    if (send_cmd(fd, GNBD_SERV_FORCE_REMOVE_REQ, "force remove") < 0)
      return 1;
  }
  else{
    if (send_cmd(fd, GNBD_SERV_REMOVE_REQ, "remove") < 0)
      return 1;
  }
  if (write(fd, &remove_req, sizeof(remove_req)) != sizeof(remove_req)){
    printe("sending remove req failed: %s\n", strerror(errno));
    return 1;
  }
  if (recv_reply(fd, (force)? "force_remove" : "remove") < 0)
    return 1;
  printm("removed GNBD %s\n", name);
  close(fd);
  return 0;
}

int validate(void){
  int fd;
  
  fd = connect_to_gnbd_servd();
  if (fd < 0)
    return 1;
  if (send_cmd(fd, GNBD_SERV_TEST_ACTIVE_REQ, "validate") < 0)
    return 1;
  if (recv_reply(fd, "validate") < 0)
    return 1;
  printm("validated all gserv processes\n");
  close(fd);
  return 0;
}

int get_list_info(void **buffer, int *buffer_size){
  void *buf;
  u32 size;
  int n, total;
  int fd;

  *buffer = NULL;
  *buffer_size = 0;
  
  fd = connect_to_gnbd_servd();
  if (fd < 0)
    return -1;
  if (send_cmd(fd, GNBD_SERV_FULL_LIST_REQ, "list") < 0)
    return -1;
  if (recv_reply(fd, "list") < 0)
    return -1;
  if (read(fd, &size, sizeof(size)) != sizeof(size)){
    printe("receiving list size failed : %s\n", strerror(errno));
    return -1;
  }
  size = be32_to_cpu(size);
  if (size == 0){
    *buffer_size = size;
    close(fd);
    return 0;
  }
  buf = malloc(size);
  if (!buf){
    printe("couldn't allocate memory for list : %s\n", strerror(errno));
    return -1;
  }
  total = 0;
  while(total < size){
    n = read(fd, buf + total, size - total);
    if (n <= 0){
      printe("receiving list failed : %s\n", strerror(errno));
      free(buf);
      return -1;
    }
    total += n;
  }

  *buffer = buf;
  *buffer_size = size;
  close(fd);
  return 0;
}


int removeall(int force){
  int size;
  void *buf;
  full_list_t *info;

  if (get_list_info(&buf, &size) < 0)
    return 1;
  if (size == 0)
    return 0;
  info = (full_list_t *)buf;
  while ((void *)info < buf + size){
    FULL_LIST_TO_CPU(info);
    if (servremove(SERVER_NAME(info), force) != 0){
      free(buf);
      return 1;
    }
    info = NEXT_SERVER(info);
  }
  free(buf);
  return 0;
}

int list(void){
  int size;
  void *buf;
  full_list_t *info;
  int i = 0;
  
  if (get_list_info(&buf, &size) < 0)
    return 1;
  if (message_flag == 0)
    return 0; 
  if (size == 0){
    printm("no exported GNBDs\n");
    free(buf);
    return 0;
  }
  info = (full_list_t *)buf;
  while ((void *)info < buf + size){
    i++;
    FULL_LIST_TO_CPU(info);
    printf("Server[%d] : %s\n"
           "--------------------------\n"
           "      file : %s\n"
           "exportsize : %"PRIu64"\n"
           " blocksize : %u\n"
           "  readonly : %s\n"
           "    cached : %s\n",
           i, SERVER_NAME(info), DEV_NAME(info), info->devsize,
           info->blksize, (info->readonly)? "yes" : "no",
           (info->timeout)? "no" : "yes");
    if (info->timeout)
      printf("   timeout : %u\n", info->timeout);
    else
      printf("   timeout : no\n");
    printf("\n");
    info = NEXT_SERVER(info);
  }
  free(buf);
  return 0;
}

int usage(void){
  printf(
"Usage:\n"
"\n"
"gnbd_export [options]\n"
"\n"
"Options:\n"
"  -a               validate the gserv processes and remove invalid ones\n"
"  -c               enable caching (used with -e)\n"
"  -d <device>      the device to create a GNBD on\n"
"  -e <GNBD>        export the specified GNBD\n"
"  -h               print this help message\n"
"  -l               list the exported GNBDS (default)\n"
"  -O               Force unexport. (used with -r and -R)\n"
"  -o               export device readonly (used with -e)\n"
"  -p <port>        specify an alternate port for gnbd_servd\n"
"  -q               quiet mode\n"
"  -R               unexport all GNBDs\n"
"  -r [GNBD | list] unexport the specified GNBD(s)\n"
"  -t <seconds>     set the timeout duration (default 60)\n"              
"  -v               verbose output (useful with -l)\n"
"  -V               version information\n");
  return 0;
}


#define ACTION_EXPORT     1
#define ACTION_REMOVE     2
#define ACTION_LIST       3
#define ACTION_REMOVE_ALL 4
#define ACTION_VALIDATE   5

char action_to_flag(int action){
  switch(action){
  case ACTION_EXPORT:
    return 'e';
  case ACTION_REMOVE:
    return 'r';
  case ACTION_LIST:
    return 'l';
  case ACTION_REMOVE_ALL:
    return 'R';
  case ACTION_VALIDATE:
    return 'a';
  default:
    printe("invalid action value\n");
    return 0;
  }
}

#define set_action(x) \
do{ \
  if (action){ \
    printe("flags -%c and -%c are not compatible\n", action_to_flag(action), \
           action_to_flag(x)); \
    fprintf(stderr, "Please see man page for details.\n"); \
    return 1; \
  } \
  action = (x); \
} while(0)

int main(int argc, char **argv){
  int c, i;
  int action = 0;
  int cached = 0;
  unsigned int timeout = 0;
  int force = 0;
  int readonly = 0;
  char *device = NULL;
  char *gnbd_name = NULL;

  while ((c = getopt(argc, argv, "acd:e:hlOop:qrRt:vV")) != -1){
    switch(c){
    case ':':
    case '?':
      fprintf(stderr, "Please use '-h' for usage.\n");
      return 1;
    case 'a':
      set_action(ACTION_VALIDATE);
      continue;
    case 'c':
      cached = 1;
      continue;
    case 'd':
      device = optarg;
      continue;
    case 'e':
      set_action(ACTION_EXPORT);
      gnbd_name = optarg;
      continue;
    case 'h':
      return usage();
    case 'l':
      set_action(ACTION_LIST);
      continue;
    case 'O':
      force = 1;
      continue;
    case 'o':
      readonly = 1;
      continue;
    case 'p':
      {
        unsigned int tmp;
        if (sscanf(optarg, "%u", &tmp) != 1 || tmp > 65535){
          printe("invalid port %s with -p\n" MAN_MSG, optarg);        
          return 1;
        }
        port = (u16)tmp;
        continue;
      }
    case 'q':
      message_flag = 0;
      continue;
    case 'r':
      set_action(ACTION_REMOVE);
      continue;
    case 'R':
      set_action(ACTION_REMOVE_ALL);
      continue;
    case 't':
      if (sscanf(optarg, "%u", &timeout) != 1 || timeout == 0){
        printe("invalid timeout '%s' with -t\n" MAN_MSG, optarg);
        return 1;
      }
      continue;
    case 'v':
      message_flag = 2;
      continue;
    case 'V':
      printf("%s %s (built %s %s)\n", argv[0],
             GNBD_RELEASE_NAME, __DATE__, __TIME__);
      printf("%s\n", SISTINA_COPYRIGHT);
      return 0;
    default:
      printe("invalid option -- %c\n", c);
      printf("Please use '-h' for usage.\n");
      return 1;
    }
  }
  if (timeout && cached){
    printe("the -t option may not be used with the -c option\n" MAN_MSG);
    return 1;
  }
  if ((cached || timeout || device || readonly) && action != ACTION_EXPORT){
    printe("the -c, -t, and -d flags may only be used with -e\n" MAN_MSG);
    return 1;
  }
  if (force && action != ACTION_REMOVE && action != ACTION_REMOVE_ALL){
    printe("the -O option mhy only be used with -r or -R.\n" MAN_MSG);
    return 1;
  }
  if (action != ACTION_REMOVE && optind != argc){
    printe("extra operand for action: %s\n", argv[optind]);
    fprintf(stderr, "please use '-h' for usage.\n");
    return 1;
  }
  switch(action){
  case ACTION_EXPORT:
    if (!device){
      printe("The -d option must be specified with -e.\n" MAN_MSG);
      return 1;
    }
    if (cached == 0 && timeout == 0)
      timeout = TIMEOUT_DEFAULT;
    return servcreate(gnbd_name, device, (u32)timeout, (u8)readonly);
  case ACTION_REMOVE:
    if (argc == optind){
      printe("missing operand for remove action\n");
      fprintf(stderr, "please use '-h' for usage.\n");
      return 1;
    }
    for (i = optind; i < argc; i++){
      if (servremove(argv[i], force) != 0)
        return 1;
    }
    return 0;
  case ACTION_LIST: case 0:
    return list();
  case ACTION_REMOVE_ALL:
    return removeall(force);
  case ACTION_VALIDATE:
    return validate();
  default:
    printe("unrecognized action value\n");
    return 1;
  }
}

